import os.path as osp

import torch
import wandb
from io import StringIO
import jammy.io as io
from jammy.logging import get_logger
from jammy.utils.printing import stprint


# TODO: remove this file!!!

__all__ = ["load_checkpoint", "save_checkpoint", "checkpoint_state"]

logger = get_logger()


def wandb_trainer_update(stage, record_dict, step_cnt):
    book = {f"{stage}/{k}": v for k, v in record_dict.items()}
    wandb.log(book, step=step_cnt)


def load_checkpoint(model=None, optimizer=None, filename="checkpoint"):
    filename = "{}.pth.tar".format(filename)

    if osp.isfile(filename):
        logger.critical("==> Loading from checkpoint '{}'".format(filename))
        checkpoint = torch.load(filename)
        env = None
        if "env" in checkpoint and checkpoint["env"] is not None:
            env = checkpoint["env"]
            mem_buffer = StringIO()
            stprint(env, file=mem_buffer)
            logger.info("\n" + mem_buffer.getvalue())
        else:
            logger.critical("No trainer env find")
        if model is not None and checkpoint["model_state"] is not None:
            model.load_state_dict(checkpoint["model_state"])
        if optimizer is not None and checkpoint["optimizer_state"] is not None:
            optimizer.load_state_dict(checkpoint["optimizer_state"])
        logger.critical("==> Loading ckpt Done")
        return env
    else:
        logger.critical("==> Checkpoint '{}' not found".format(filename))
        return None


def save_checkpoint(state, is_best, filename="checkpoint", bestname="model_best"):
    filename = "{}.pth.tar".format(filename)
    torch.save(state, filename)
    if is_best:
        io.copy(filename, "{}.pth.tar".format(bestname))
        logger.info("Save best up to now: {}".format(filename))


def checkpoint_state(model=None, optimizer=None):
    optim_state = optimizer.state_dict() if optimizer is not None else None
    if model is not None:
        if isinstance(model, torch.nn.DataParallel):
            model_state = model.module.state_dict()
        else:
            model_state = model.state_dict()
    else:
        model_state = None

    return {
        "model_state": model_state,
        "optimizer_state": optim_state,
    }



from omegaconf import OmegaConf
import os.path as osp
import hydra

__all__ = ["hydpath"]


def hydpath(input_path=None):
    if input_path is None or len(input_path) == 0:
        return hydra.utils.to_absolute_path(".")
    if input_path[0] == "/":
        return input_path
    if input_path[0] == "~":
        return osp.expanduser(input_path)
    return hydra.utils.to_absolute_path(input_path)



import time

import pytorch_lightning as pl  # pylint: disable=unused-import
import torch
from pytorch_lightning import Callback
from pytorch_lightning.utilities import rank_zero_info

# pylint: disable=attribute-defined-outside-init


class CUDACallback(Callback):
    def on_train_epoch_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        del pl_module
        # Reset the memory use counter
        torch.cuda.reset_peak_memory_stats(trainer.root_gpu)
        torch.cuda.synchronize(trainer.root_gpu)
        self.start_time = time.time()

    def on_train_epoch_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        torch.cuda.synchronize(trainer.root_gpu)
        max_memory = torch.cuda.max_memory_allocated(trainer.root_gpu) / 2 ** 20
        epoch_time = time.time() - self.start_time

        try:
            max_memory = trainer.training_type_plugin.reduce(max_memory)
            epoch_time = trainer.training_type_plugin.reduce(epoch_time)

            pl_module.log_dict(
                {
                    "profile/epoch_time (sec)": epoch_time,
                    "profile/peak_mem (MB)": max_memory,
                }
            )
            rank_zero_info(f"Average Epoch time: {epoch_time:.2f} seconds")
            rank_zero_info(f"Average Peak memory {max_memory:.2f}MiB")
        except AttributeError:
            pass



import numpy as np
import scipy.misc
# FIXME: bug for tf1,tf2
try:
    from StringIO import StringIO as BytesIO  # Python 2.7
except ImportError:
    from io import BytesIO  # Python 3.x

import tensorflow

if tensorflow.__version__ >= "1.14.0":
    import tensorflow.compat.v1 as tf
    create_fn = tensorflow.summary.create_file_writer
else:
    import tensorflow as tf
    create_fn = tf.summary.FileWriter


class TBLogger(object):
    # Adapted from:
    # https://raw.githubusercontent.com/SherlockLiao/pytorch-beginner/

    def __init__(self, log_dir):
        self.writer = create_fn(log_dir)

    def scalar_summary(self, tag, value, step):
        summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
        self.writer.add_summary(summary, step)

    def image_summary(self, tag, images, step):
        img_summaries = []
        for i, img in enumerate(images):
            # BUG: remove toimage
            # Write the image to a string
            s = BytesIO()
            scipy.misc.toimage(img).save(s, format="png")

            # Create an Image object
            img_sum = tf.Summary.Image(
                encoded_image_string=s.getvalue(),
                height=img.shape[0],
                width=img.shape[1],
            )
            # Create a Summary value
            img_summaries.append(
                tf.Summary.Value(tag="%s/%d" % (tag, i), image=img_sum)
            )

        # Create and write Summary
        summary = tf.Summary(value=img_summaries)
        self.writer.add_summary(summary, step)

    def histo_summary(self, tag, values, step, bins=1000):
        # Create a histogram using numpy
        counts, bin_edges = np.histogram(values, bins=bins)

        # Fill the fields of the histogram proto
        hist = tf.HistogramProto()
        hist.min = float(np.min(values))
        hist.max = float(np.max(values))
        hist.num = int(np.prod(values.shape))
        hist.sum = float(np.sum(values))
        hist.sum_squares = float(np.sum(values ** 2))

        # Drop the start of the first bin
        bin_edges = bin_edges[1:]

        # Add bin edges and counts
        for edge in bin_edges:
            hist.bucket_limit.append(edge)
        for c in counts:
            hist.bucket.append(c)

        # Create and write Summary
        summary = tf.Summary(value=[tf.Summary.Value(tag=tag, histo=hist)])
        self.writer.add_summary(summary, step)

    def flush(self):
        self.writer.flush()



import torch
import numpy as np

__all__ = [
    "FloatTensor",
    "from_numpy",
    "get_numpy",
    "gpu_enabled",
    "normal",
    "ones",
    "ones_like",
    "randn",
    "set_device",
    "set_gpu_mode",
    "soft_update_from_to",
    "tensor",
    "zeros",
    "zeros_like",
    "device",
]


def soft_update_from_to(source, target, tau):
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)


def copy_model_params_from_to(source, target):
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(param.data)


def fanin_init(tensor):
    size = tensor.size()
    if len(size) == 2:
        fan_in = size[0]
    elif len(size) > 2:
        fan_in = np.prod(size[1:])
    else:
        raise Exception("Shape must be have dimension at least 2.")
    bound = 1.0 / np.sqrt(fan_in)
    return tensor.data.uniform_(-bound, bound)


def fanin_init_weights_like(tensor):
    size = tensor.size()
    if len(size) == 2:
        fan_in = size[0]
    elif len(size) > 2:
        fan_in = np.prod(size[1:])
    else:
        raise Exception("Shape must be have dimension at least 2.")
    bound = 1.0 / np.sqrt(fan_in)
    new_tensor = FloatTensor(tensor.size())
    new_tensor.uniform_(-bound, bound)
    return new_tensor


"""
GPU wrappers
"""

_use_gpu = False
device = torch.device("cpu")
_gpu_id = 0


def set_gpu_mode(is_gpu=True, gpu_id=0):
    """set jam pytorch utils device

    :param is_gpu: [description], defaults to True
    :type is_gpu: bool, optional
    :param gpu_id: [description], defaults to 0
    :type gpu_id: int, optional
    """
    global _use_gpu
    global device
    global _gpu_id
    _gpu_id = gpu_id
    _use_gpu = is_gpu
    device = torch.device("cuda:" + str(gpu_id) if _use_gpu else "cpu")


def gpu_enabled():
    return _use_gpu


def set_device(gpu_id):
    torch.cuda.set_device(gpu_id)


# noinspection PyPep8Naming
def FloatTensor(*args, torch_device=None, **kwargs):
    if torch_device is None:
        torch_device = device
    return torch.FloatTensor(*args, **kwargs).to(device)


def from_numpy(*args, **kwargs):
    return torch.from_numpy(*args, **kwargs).float().to(device)


def get_numpy(tensor):
    return tensor.to("cpu").detach().numpy()


def zeros(*sizes, torch_device=None, **kwargs):
    if torch_device is None:
        torch_device = device
    return torch.zeros(*sizes, **kwargs, device=torch_device)


def ones(*sizes, torch_device=None, **kwargs):
    if torch_device is None:
        torch_device = device
    return torch.ones(*sizes, **kwargs, device=torch_device)


def ones_like(*args, torch_device=None, **kwargs):
    if torch_device is None:
        torch_device = device
    return torch.ones_like(*args, **kwargs, device=torch_device)


def randn(*args, torch_device=None, **kwargs):
    if torch_device is None:
        torch_device = device
    return torch.randn(*args, **kwargs, device=torch_device)


def zeros_like(*args, torch_device=None, **kwargs):
    if torch_device is None:
        torch_device = device
    return torch.zeros_like(*args, **kwargs, device=torch_device)


def tensor(*args, torch_device=None, **kwargs):
    if torch_device is None:
        torch_device = device
    return torch.tensor(*args, **kwargs, device=torch_device)


def normal(*args, **kwargs):
    return torch.normal(*args, **kwargs).to(device)

def randn(*args, **kwargs):
    return torch.randn(*args, **kwargs).to(device)



from jammy.utils.hyd import hyd_instantiate

from jamtorch.logging import get_logger

logger = get_logger()

__all__ = ["hyd_ema"]

def hyd_ema(cfg):
    is_ema = False
    if "enable_ema" in cfg:
        if cfg.enable_ema != True:
            return None
    if "ema" in cfg:
        if bool(cfg.ema):
            return hyd_instantiate(cfg.ema)
        if "enable_ema" in cfg and cfg.enable_ema == True:
            logger.critical("ema fail instantiate")
    return None




from .simple_trainer import SimpleTrainer

import jammy.utils.hyd as hyd

class EMATrainer(SimpleTrainer):
    def __init__(self, cfg, loss_fn):
        super().__init__(cfg, loss_fn)
        self.ema = hyd.hyd_instantiate(cfg.ema)
        if self.ema:
            def update_ema(trainer):
                trainer.ema.update_parameters(trainer.model)
            self.register_event("step:end",update_ema, False)

    def _impl_load_ckpt(self, state):
        if self.ema:
            self.ema.load_state_dict(state["ema"])
        return super()._impl_load_ckpt(state)

    def _impl_save_ckpt(self):
        rtn = {}
        if self.ema:
            rtn["ema"] = self.ema.state_dict()
        return {**super()._impl_save_ckpt(), **rtn}


#!/usr/bin/env python3
# pylint: disable=global-variable-not-assigned, global-statement
import functools
import os
import os.path as osp
import re
from subprocess import Popen

import hydra
from omegaconf import DictConfig, OmegaConf

from jammy.logging import get_logger

logger = get_logger()

jam_dk_wrappers = []
dk_args = ["docker", "run", "-d", "--gpus", "all", "--privileged", "-it"]
print_args = ["\n", "\t".join(dk_args)]


def jamdk_wrapper(func):
    global jam_dk_wrappers, dk_args, print_args

    @functools.wraps(func)
    def _wrapped_func(*args, **kwargs):
        rtn = func(*args, **kwargs)
        if rtn is not None:
            dk_args.extend(rtn)
            print_args.append("\t".join(rtn))
        return dk_args

    jam_dk_wrappers.append(_wrapped_func)
    return _wrapped_func


def read_local_cfg(cfg):
    cfg_path = hydra.utils.to_absolute_path(cfg.file)
    if not osp.exists(cfg_path):
        return cfg
    logger.info("readding config from: " + cfg_path)
    local_cfg = OmegaConf.load(cfg_path)
    cli_cfg = OmegaConf.from_cli()
    cfg = OmegaConf.merge(cfg, local_cfg, cli_cfg)
    return cfg


@jamdk_wrapper
def check_network(cfg):
    if cfg.network is None:
        return ["--network", "host"]
    return ["--network", cfg.network]


@jamdk_wrapper
def set_env(cfg):
    if "env" in cfg:
        if cfg.env is not None:
            regexp = re.compile(r"^\$[a-zA-Z0-9_]*$")
            rtn = []
            for key, value in cfg.env.items():
                if regexp.search(value):
                    if value[1:] in os.environ:
                        value = os.environ[value[1:]]
                rtn.extend(["--env", f"{key}={value}"])
            return rtn
    return None


@jamdk_wrapper
def check_memeory(cfg):
    mem_bytes = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
    mem_gib = mem_bytes / (1024.0**3)
    if cfg.memory is None:
        cfg.memory = int(mem_gib)
    return ["--shm-size", f"{cfg.memory}G"]


@jamdk_wrapper
def check_x_display(cfg):
    display = os.getenv("DISPLAY", None)
    if display is None or cfg.x is False:
        return None
    return [
        "--volume=/tmp/.X11-unix:/tmp/.X11-unix",
        "--volume=/tmp/.docker.xauth:/tmp/.docker.xauth",
        "-e",
        f"DISPLAY={display}",
    ]


@jamdk_wrapper
def check_ssh_mount(cfg):
    if cfg.ssh:
        ssh_path = osp.expanduser("~/.ssh/")
        return [f"--volume={ssh_path}:{ssh_path}:rw"]
    return None


@jamdk_wrapper
def check_jam_mount(cfg):
    if cfg.jam:
        jam_path = osp.expanduser("~/jam/")
        return [f"--volume={jam_path}:{jam_path}:rw"]
    return None


@jamdk_wrapper
def extra_mount(cfg):
    if cfg.mount is not None:
        rtn = []
        for local_f, dk_f in cfg.mount.items():
            if local_f[0] == "~":
                local_f = osp.expanduser(local_f)
            if local_f[0] != "/":
                local_f = hydra.utils.to_absolute_path(local_f)
            if dk_f is None:
                dk_f = local_f
            else:
                if dk_f[0] == "~":
                    dk_f = osp.expanduser(dk_f)
                if dk_f[0] != "/":
                    dk_f = hydra.utils.to_absolute_path(dk_f)
            rtn.append(f"--volume={local_f}:{dk_f}:rw")
        return rtn
    return None


@jamdk_wrapper
def extra_aug(cfg):
    if "extra" in cfg:
        if cfg.extra is not None:
            rtn = []
            for key, value in cfg.extra.items():
                rtn.append(f"--{key}={value}")
            return rtn
    return None


@jamdk_wrapper
def container_name(cfg):
    if cfg.name is not None:
        return ["--name", cfg.name]
    return None


@jamdk_wrapper
def img(cfg):
    if cfg.img is not None:
        return [f"{cfg.img}"]
    raise RuntimeError


@hydra.main(config_path="conf", config_name="jamdk", version_base="1.1.0")
def my_app(cfg: DictConfig) -> None:
    OmegaConf.set_struct(cfg, False)
    cfg = read_local_cfg(cfg)
    for item in jam_dk_wrappers:
        item(cfg)
    dk_args.extend("tail -f /dev/null".split())
    logger.debug(dk_args)
    logger.info(f"stating {cfg.name}" + "\n".join(print_args))
    Popen(dk_args)  # pylint: disable=consider-using-with


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : argument.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/18/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.

import os.path as osp
import argparse
import json

from jammy.logging import get_logger
from jammy.utils.enum import JamEnum
from jammy.utils.printing import print_to
from jammy.utils.container import G

from .device import DeviceNameFormat, parse_and_set_devices
from .keyboard import str2bool, yn2bool

__all__ = ["JamArgumentParser"]

logger = get_logger()


class JamArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("fromfile_prefix_chars", "@")
        kwargs.setdefault("formatter_class", argparse.ArgumentDefaultsHelpFormatter)
        super().__init__(*args, **kwargs)

        self.register("type", "bool", _type_bool)
        self.register("type", "checked_file", _type_checked_file)
        self.register("type", "checked_dir", _type_checked_dir)
        self.register("type", "ensured_dir", _type_ensured_dir)
        self.register("type", "kv", _type_kv)
        self.register("action", "set_device", SetDeviceAction)
        self.register("action", "as_enum", AsEnumAction)


def _type_bool(string):
    try:
        return str2bool(string)
    except ValueError:
        raise argparse.ArgumentTypeError()


def _type_checked_file(string):
    if not osp.isfile(string):
        raise argparse.ArgumentTypeError(
            'Check file existence failed: "{}".'.format(string)
        )
    return string


def _type_checked_dir(string):
    if not osp.isdir(string):
        raise argparse.ArgumentTypeError(
            'Check directory existence failed: "{}".'.format(string)
        )
    return string


def _type_ensured_dir(string):
    if not osp.isdir(string):
        import jammy.io as io
        from jammy.cli import yes_or_no

        if yes_or_no(
            f"dir {string} is missing, continue before creating an empty folder?"
        ):
            io.mkdir(string)
    return string


class _KV(object):
    def __init__(self, string):
        self.string = string

        if len(self.string) > 0:
            kvs = list(string.split(";"))
        else:
            kvs = []

        for i, kv in enumerate(kvs):
            k, v = kv.split("=")

            try:
                bool_value = yn2bool(v)
                v = bool_value
                kvs[i] = (k, v)
                continue
            except ValueError:
                pass

            if v.startswith('"') or v.startswith("'"):
                assert v.endswith('"') or v.endswith("'")
                v = v[1:-1]

            try:
                v = float(v)
                if int(v) == v:
                    v = int(v)
            except:
                pass

            kvs[i] = (k, v)

        self.kvs = kvs

    def apply(self, configs):
        with print_to(logger.info):
            print("Applying KVs:")
            for k, v in self.kvs:
                print("  kv.{} = {}".format(k, v))
                keys = k.split(".")
                current = configs
                for k in keys[:-1]:
                    current = current.setdefault(k, G())
                current[keys[-1]] = v

    def __jsonify__(self):
        return json.dumps(self.kvs)


def _type_kv(string):
    """
    In the format of:
        --configs "data.int_or_float=int_value; data.string='string_value'"
    """

    return _KV(string)


class SetDeviceAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest,
        format="int",
        set_device=True,
        nargs=None,
        const=None,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):

        DeviceNameFormat.assert_valid(format)
        self.format = format
        self.set_device = set_device

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(
            namespace,
            self.dest,
            parse_and_set_devices(values, self.format, self.set_device),
        )


class AsEnumAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest,
        type,
        nargs=None,
        const=None,
        default=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):

        assert issubclass(type, JamEnum)

        self.enum_type = type
        if choices is None:
            choices = type.choice_values()
        if default is not None:
            default = self.enum_type.from_string(default)

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=None,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, (tuple, list)):
            setattr(
                namespace, self.dest, tuple(map(self.enum_type.from_string, values))
            )
        else:
            setattr(namespace, self.dest, self.enum_type.from_string(values))



import sys

from loguru import logger

__all__ = [
    "get_logger",
    "LOG_FORMAT",
]

logger.remove()

logger_sink = {}
LOG_FORMAT = (
    "[<green>{time:MM-DD HH:mm:ss}</green>|"
    "<red>{process.name}</red>|"
    "<level>{level: <8}</level>|"
    "<cyan>{file.path}</cyan>:<cyan>{line}</cyan>:<cyan>{function}</cyan>]"
    "<level>{message}</level>"
)


def get_logger(file_name=None, clear=False, **kwargs):
    global logger_sink  # pylint: disable=global-statement,global-variable-not-assigned
    if clear:
        logger.remove()
        logger_sink.clear()
    if file_name is None:
        file_name = sys.stderr
        if "level" not in kwargs:
            kwargs["level"] = "INFO"
    if "format" not in kwargs:
        kwargs["format"] = LOG_FORMAT
    if file_name in logger_sink.values():
        # FIXME:
        # logger.debug("already registered")
        logger.debug(f"{str(file_name)} already registered")
    else:
        # if "level" not in kwargs:
        # kwargs["level"] = "DEBUG" if jam_is_debug() else "INFO"
        sink_id = logger.add(file_name, **kwargs)
        logger_sink[sink_id] = file_name
        logger.debug(f"sink_id: {sink_id:02d} ---> {str(file_name)}")
    return logger



__all__ = [
    "is_bool",
    "is_int",
    "is_positive_int",
    "is_nonnegative_int",
    "is_power_of_two"
]

def is_bool(x):
    return isinstance(x, bool)


def is_int(x):
    return isinstance(x, int)


def is_positive_int(x):
    return is_int(x) and x > 0


def is_nonnegative_int(x):
    return is_int(x) and x >= 0


def is_power_of_two(n):
    if is_positive_int(n):
        return not n & (n - 1)
    else:
        return False



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : codecs.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

from .backend import cv2, opencv_only
import numpy as np

__all__ = ["jpeg_encode", "png_encode", "imdecode"]


@opencv_only
def jpeg_encode(img, quality=90):
    """Encode the image with JPEG encoder.

    Args:
        img (:class:`numpy.ndarray`): uint8 color image array
        quality (int): quality for JPEG compression

    Returns bytes: encoded image data
    """
    return cv2.imencode(".jpg", img, [int(cv2.IMWRITE_JPEG_QUALITY), quality])[
        1
    ].tostring()


@opencv_only
def png_encode(input, compress_level=3):
    """
    Encode the image with PNG encoder.

    Args:
        img (:class:`numpy.ndarray`): uint8 color image array
        quality (int): quality for JPEG compression

    Returns bytes: encoded image data
    """
    assert len(input.shape) == 3 and input.shape[2] in [3, 4]
    assert input.dtype == np.uint8
    assert isinstance(compress_level, int) and 0 <= compress_level <= 9
    enc = cv2.imencode(
        ".png", input, [int(cv2.IMWRITE_PNG_COMPRESSION), compress_level]
    )
    return enc[1].tostring()


@opencv_only
def imdecode(data, *, require_chl3=True, require_alpha=False):
    """Decode images in common formats (jpg, png, etc.).

    Args:
        data (bytes): encoded image data
        require_chl3: whether to convert gray image to 3-channel BGR image
        require_alpha: whether to add alpha channel to BGR image

    Returns: uint8 color image array
    """

    img = cv2.imdecode(np.fromstring(data, np.uint8), cv2.IMREAD_UNCHANGED)
    assert img is not None, "failed to decode"
    if img.ndim == 2 and require_chl3:
        img = img.reshape(img.shape + (1,))
    if img.shape[2] == 1 and require_chl3:
        img = np.tile(img, (1, 1, 3))
    if img.ndim == 3 and img.shape[2] == 3 and require_alpha:
        assert img.dtype == np.uint8
        img = np.concatenate([img, np.ones_like(img[:, :, :1]) * 255], axis=2)
    return img



import argparse
import os
import signal
import socket
import threading
import time
import uuid
from collections import defaultdict
from dataclasses import asdict, dataclass, field
from subprocess import Popen
from typing import List, Optional, Union

import gpustat
import psutil

from jammy.comm import get_local_addr, is_port_used
from jammy.comm.cs import ClientPipe, ServerPipe
from jammy.logging import get_logger
from jammy.utils.env import jam_getenv
from jammy.utils.printing import kvformat, stformat

logger = get_logger(
    os.path.expanduser("~/jsh.log"),
    level="DEBUG",
    rotation="10 MB",
    enqueue=True,
    filter=lambda record: "jsh" in record["extra"],
)
ERROR_FORMAT = (
    "<green>{time:MM-DD HH:mm:ss.SSS}</green> |"
    "<red>[{process.name}]</red>|"
    "<level>{level: <8}</level> |"
    "<level>{message}</level>"
)
logger = get_logger(
    os.path.expanduser("~/jsh_error.log"),
    level="DEBUG",
    rotation="10 MB",
    enqueue=True,
    filter=lambda record: "jsherror" in record["extra"],
    format=ERROR_FORMAT,
)

# pylint: disable=global-variable-not-assigned, global-statement


@dataclass
class PCState:
    mem_avail: float = 0.0
    cpu_usage: float = 1.0

    def ready(self, cpu_usage_thres, mem_avail_thres):
        cpu_ready = cpu_usage_thres > self.cpu_usage
        mem_ready = mem_avail_thres < self.mem_avail
        return cpu_ready and mem_ready


def get_pc_state(interval_cpu_check: int = 2):
    mem_avail = psutil.virtual_memory().available / 1024 / 1024 / 1024
    cpu_usage = psutil.cpu_percent(interval=interval_cpu_check)
    return PCState(mem_avail, cpu_usage)


@dataclass
class GPUState:
    query: gpustat.core.GPUStatCollection

    def ready(self, i_th: int, gpu_usage_thres, mem_avail_thres):
        gpu_info = self.query[i_th]
        gpu_util_ready = gpu_usage_thres > gpu_info["utilization.gpu"]
        gpu_mem_ready = (
            mem_avail_thres
            < (gpu_info["memory.total"] - gpu_info["memory.used"]) / 1024.0
        )
        return gpu_util_ready and gpu_mem_ready


@dataclass
class ProcTask:
    cmd: str
    proc: Popen
    gpus: list
    pid: int = field(init=False)

    def __post_init__(self):
        self.pid = self.proc.pid

    def poll(self):
        return self.proc.poll()


class ThreadSafeDict(dict):
    def __init__(self, *p_arg, **n_arg):
        dict.__init__(self, *p_arg, **n_arg)
        self._lock = threading.Lock()

    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, tp, value, traceback):  # pylint: disable=invalid-name
        del tp, value, traceback
        self._lock.release()


class ThreadSafeList(list):
    def __init__(self, *p_arg, **n_arg):
        list.__init__(self, *p_arg, **n_arg)
        self._lock = threading.Lock()

    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, type_p, value, traceback):
        del type_p, value, traceback
        self._lock.release()


class CmdExecutor:
    def __init__(self):
        self.active_tasks = ThreadSafeDict()
        self.run = False
        self.thread = None
        self._last_msg = []
        self._last_msg_lock = threading.Lock()

    def append(self, item, gpus: Optional[Union[None, int, list]] = None):
        # https://jorgenmodin.net/index_html/Link---unix---Killing-a-subprocess-including-its-children-from-python---Stack-Overflow
        # os.setsid assign a anew process group
        gpu_prefix = ""
        if gpus:
            gpu_prefix = "CUDA_VISIBLE_DEVICES=" + ",".join(map(str, gpus)) + " "
        str_cmd = gpu_prefix + item
        process = (
            Popen(  # pylint: disable=consider-using-with, subprocess-popen-preexec-fn
                str_cmd, shell=True, preexec_fn=os.setsid
            )
        )
        cur_task = ProcTask(str_cmd, process, gpus=gpus)
        # with Popen(item, shell=True) as process:
        with self.active_tasks as active_tasks:
            active_tasks[str_cmd] = cur_task
            return cur_task

    def __len__(self):
        return len(self.active_tasks)

    @property
    def last_msg(self) -> List[ProcTask]:
        with self._last_msg_lock:
            if len(self._last_msg) == 0:
                return None
            temp = self._last_msg
            self._last_msg = []
            return temp

    def start(self):
        self.run = True
        self.thread = threading.Thread(target=self._update_fn)
        self.thread.start()
        logger.bind(jsh=True).debug("Starting CmdExecutor")

    def stop(self):
        if self.run:
            self.run = False
            logger.bind(jsh=True).info("STOP CmdExecutor")
            with self.active_tasks as active_task:
                for cmd, task_process in active_task.items():
                    os.killpg(os.getpgid(task_process.pid), signal.SIGKILL)
                    logger.bind(jsh=True).info(
                        f"Work Term PID:{task_process.pid}\t {cmd}"
                    )
                active_task.clear()
            self.thread.join()
            self.thread = None
            with self.active_tasks as active_task:
                for cmd, task_process in active_task.items():
                    task_process.terminate()
                    logger.bind(jsh=True).info(
                        f"Work Term PID:{task_process.pid}\t {cmd}"
                    )
                active_task.clear()

    def join(self):
        if self.run:
            while len(self) > 0:
                time.sleep(3)
            self.stop()

    def _update_fn(self):
        while self.run:
            self._check_cmds()
            time.sleep(2)

    def _check_one_cmd(self, cmd: str, proc_task: ProcTask):
        # TODO: we actually do not need dict for task
        del cmd
        poll_state = proc_task.proc.poll()
        if poll_state is None:
            return False
        with self._last_msg_lock:
            self._last_msg.append(proc_task)
        return True

    def _check_cmds(self):
        del_tasks = []
        for cmd, proc_task in self.active_tasks.items():
            if self._check_one_cmd(cmd, proc_task):
                del_tasks.append(cmd)
        with self.active_tasks as active_tasks:
            for key in del_tasks:
                del active_tasks[key]

    def __str__(self):
        return "Active tasks\n" + stformat(
            {proc.pid: proc.cmd for proc in self.active_tasks.values()}
        )


@dataclass
class SchedulerSetting:  # pylint: disable=too-many-instance-attributes
    seconds_resouece_wait: float = 2.0
    interval_cpu_check: int = 1
    cpu_usage_thres: float = 92
    mem_avail_thres: float = 0.9
    cpu_proc_upper: int = 4

    gpu_proc_upper: int = 2
    gpu_usage_thres: float = 80
    gpu_mem_avail_thres: float = 0.4


class Scheduler:  # pylint: disable=too-many-instance-attributes
    def __init__(self):
        self._cmd_executor = CmdExecutor()
        self._task_buffer = ThreadSafeList()

        self.num_total_task = 0
        self.num_finished_task = 0
        self.num_error_task = 0

        self.thread = None
        self.run = False

        self.cfg = SchedulerSetting()

        # gpus
        self.gpus_num_proc = defaultdict(lambda: 0)

    def reset(self):
        self._cmd_executor = CmdExecutor()
        self._task_buffer = ThreadSafeList()

        self.num_total_task = 0
        self.num_finished_task = 0
        self.num_error_task = 0

        self.thread = None
        self.run = False

    def start(self):
        self.run = True
        self._cmd_executor.start()
        self.thread = threading.Thread(target=self._update_fn)
        self.thread.start()

    def stop(self):
        if self.run:
            logger.bind(jsh=True).info("STOP Scheduler")
            self.run = False
            self.thread.join()
            self.thread = None
            self._cmd_executor.stop()

    def join(self):
        if self.run:
            while len(self._task_buffer) > 0:
                time.sleep(3)
            self._cmd_executor.join()
        self.stop()

    def append(self, cmd: str, gpus: Optional[Union[None, int, list]] = None):
        with self._task_buffer as buffer:
            if isinstance(gpus, int):
                gpus = [gpus]
            buffer.append((cmd, gpus))
            logger.bind(jsh=True).debug(f"Scheduler recieves {cmd} on GPUs: {gpus}")

    def _update_fn(self):
        logger.bind(jsh=True).debug("Entering Scheduler worker")
        while self.run:
            runner_msgs = self._cmd_executor.last_msg
            self._process_msg(runner_msgs)

            if len(self._task_buffer) == 0:
                time.sleep(3)
                continue

            if not self.is_cpu_ready():
                time.sleep(self.cfg.seconds_resouece_wait)
                continue

            gpu_ready = False
            cur_i = 0
            for cur_i, cur_task in enumerate(self._task_buffer):
                str_cmd, req_gpus = cur_task
                req_gpus = self.check_gpu_input(req_gpus)  # check if need auto tune gpu
                if self.is_gpu_ready(req_gpus):
                    gpu_ready = True
                    break
            if gpu_ready:
                with self._task_buffer as buffer:
                    del buffer[cur_i]

                proc_task = self._cmd_executor.append(str_cmd, req_gpus)
                self.post_start_process(proc_task)
            else:
                time.sleep(1)
        logger.bind(jsh=True).debug("Exiting Scheduler worker")

    def event_info(self, proc_task: Union[ProcTask, None] = None):
        if proc_task:
            return f"\nPID:{proc_task.pid}. Active:{len(self._cmd_executor):>5d}\
                Finished/TODO/Error: {self.num_finished_task:>5d}/{len(self._task_buffer):>5d}/{self.num_error_task:>5d}"  # pylint: disable=line-too-long
        return f"Finished/TODO/Error: {self.num_finished_task:>5d}/{len(self._task_buffer):>5d}/{self.num_error_task:>5d}"  # pylint: disable=line-too-long

    def log_state(self):
        logger.bind(jsh=True).debug(
            f"Active:{len(self._cmd_executor)}. Finished/TODO/Error: {self.num_finished_task:>5d}/{len(self._task_buffer):>5d}/{self.num_error_task:>5d}"  # pylint: disable=line-too-long
        )

    def _process_msg(self, msgs: List[ProcTask]) -> None:
        if msgs:
            for proc_task in msgs:
                poll_state = proc_task.poll()
                if poll_state == 0:
                    self.post_work_cmd(proc_task)
                else:
                    self.post_broken_cmd(proc_task)

    def is_cpu_ready(self):
        pc_state = get_pc_state(self.cfg.interval_cpu_check)
        state = pc_state.ready(self.cfg.cpu_usage_thres, self.cfg.mem_avail_thres)
        is_lower_upper_proc = len(self._cmd_executor) < self.cfg.cpu_proc_upper
        return state and is_lower_upper_proc

    def is_gpu_ready(self, gpus: List[int] = None):
        if gpus:
            gpu_state = GPUState(gpustat.new_query())
            for cur_id in gpus:
                if self.gpus_num_proc[cur_id] >= self.cfg.gpu_proc_upper:
                    return False

                if not gpu_state.ready(
                    cur_id, self.cfg.gpu_usage_thres, self.cfg.gpu_mem_avail_thres
                ):
                    return False

        return True

    def check_gpu_input(self, gpus: List[int] = None):
        if gpus:
            if len(gpus) == 1 and gpus[-1] < 0:
                rtn_gpu = self.select_gpu(abs(gpus[-1]))
                if rtn_gpu:
                    return rtn_gpu
        return gpus

    def select_gpu(self, num_gpu: int = 2):
        rtn = []
        gpu_state = GPUState(gpustat.new_query())
        for cur_id in range(len(gpu_state.query)):
            if self.gpus_num_proc[cur_id] >= self.cfg.gpu_proc_upper:
                continue

            if not gpu_state.ready(
                cur_id, self.cfg.gpu_usage_thres, self.cfg.gpu_mem_avail_thres
            ):
                continue
            rtn.append(cur_id)
        if len(rtn) < num_gpu:
            return None
        logger.bind(jsh=True).debug(f"Select GPUs : {rtn[:num_gpu]}")
        return rtn[:num_gpu]

    def post_start_process(self, proc_task: ProcTask):
        self.num_total_task += 1
        if proc_task.gpus:
            for item in proc_task.gpus:
                self.gpus_num_proc[item] += 1
        logger.bind(jsh=True).info(
            f"\nFire {proc_task.cmd}! {self.event_info(proc_task)}"
        )

    def post_work_cmd(self, proc_task: ProcTask):
        self.num_finished_task += 1
        if proc_task.gpus:
            for item in proc_task.gpus:
                self.gpus_num_proc[item] -= 1
        logger.bind(jsh=True).info(
            f"\nFinish {proc_task.cmd}! {self.event_info(proc_task)}"
        )

    def post_broken_cmd(self, proc_task: ProcTask):
        self.num_finished_task += 1
        if proc_task.gpus:
            for item in proc_task.gpus:
                self.gpus_num_proc[item] -= 1

        poll_state = proc_task.poll()
        logger.bind(jsh=True).critical(
            f"\nError:\t{poll_state:03d}!\n{proc_task.cmd}{self.event_info(proc_task)}"
        )
        logger.bind(jsherror=True).debug(f"PID:{proc_task.pid:>06d} {proc_task.cmd}")

    def __exit__(self, *args, **kwargs):
        self.stop()

    def __str__(self):
        return (
            "jsh setting\n"
            + stformat(asdict(self.cfg))
            + "\n\n TODO\n"
            + stformat(self._task_buffer)
            + self.event_info()
            + "\n\n"
            + str(self._cmd_executor)
            + "GPUs\n"
            + stformat(dict(self.gpus_num_proc))
        )


worker = None


def set_scheduler(pipe, identifier, inp=None):
    global worker
    idx = identifier.decode("ascii")
    logger.info(f"REQ SET from {idx}\n {kvformat(inp)}")
    for key, value in inp.items():
        try:
            setattr(worker.cfg, key, value)
        except Exception as error:  # pylint: disable=broad-except
            print(error)
    logger.debug("SET Done")
    pipe.send(identifier, None)


def add_job(pipe, identifier, inp=None):
    global worker
    idx = identifier.decode("ascii")
    if isinstance(inp, (str, tuple)):
        inp = [inp]
    logger.info(f"REQ job from {idx}\n {stformat(inp)}")
    for job in inp:
        if isinstance(job, str):
            worker.append(job)
        else:
            worker.append(job[0], job[1])
    logger.debug("Add work done")
    pipe.send(identifier, None)


def kill_all(pipe, identifier, inp=None):
    del inp
    global worker
    idx = identifier.decode("ascii")
    logger.info(f"REQ kill_all from {idx}\n")
    worker.stop()
    worker.reset()
    worker.start()
    pipe.send(identifier, None)


def response_state(pipe, identifier, inp=None):
    del inp
    global worker
    idx = identifier.decode("ascii")
    logger.info(f"REQ state from {idx}\n")
    pipe.send(identifier, str(worker))


def start_sever():
    parser = argparse.ArgumentParser(prog="jshs: start a shell executor server")
    parser.add_argument("-p", "--port", type=int, default=-1, help="set default port")
    args = parser.parse_args()

    global worker
    worker = Scheduler()
    worker.start()
    try:
        server = ServerPipe("server", mode="tcp")
        server.dispatcher.register("set", set_scheduler)
        server.dispatcher.register("job", add_job)
        server.dispatcher.register("killf", kill_all)
        server.dispatcher.register("state", response_state)
        if args.port < -1:
            p_router = jam_getenv("ShExecutor", default=1089, type=int)
            logger.warning(f"Use default port {p_router}")
        else:
            p_router = args.port
            logger.warning(f"Use passed Port {p_router}")
        assert not is_port_used(p_router)
        assert not is_port_used(p_router + 1)
        with server.activate(tcp_port=[str(p_router), str(p_router + 1)]):
            logger.critical("GPU ROUTER and PULL IP:")
            logger.critical(server.conn_info)
            while True:
                time.sleep(1)
    except KeyboardInterrupt:
        worker.stop()


def instantiate_client(flag="_", p_dealer=-1):
    if p_dealer < 0:
        p_dealer = jam_getenv("ShExecutor", default=1089, type=int)
        logger.warning(f"Use default port {p_dealer}")
    tcp_dealer = f"tcp://{get_local_addr()}:{p_dealer}"
    tcp_push = f"tcp://{get_local_addr()}:{p_dealer+1}"
    client = ClientPipe(
        f"gpu_client{flag}" + uuid.uuid4().hex[:8], conn_info=[tcp_dealer, tcp_push]
    )
    return client


def _shell_instantiate_client(port=-1):
    parser = argparse.ArgumentParser(prog="jsh client")
    parser.add_argument("-p", "--port", type=int, default=-1, help="set default port")
    args = parser.parse_args()
    if args.port > 0:
        port = args.port
        logger.warning(f"Use passed Port {args.port}")
    return instantiate_client(p_dealer=port)


def echo_hello():
    client = _shell_instantiate_client()
    with client.activate():
        client.query("job", f"echo Hello from {socket.gethostname()} && pwd")


def echo_state():
    client = _shell_instantiate_client()
    with client.activate():
        state_info = client.query("state")
        print(state_info)


def client_kill_all():
    client = _shell_instantiate_client()
    logger.info("Identity: {}.".format(client.identity))
    with client.activate():
        client.query("killf")


# TODO: adding state output and fzf kill

if __name__ == "__main__":
    worker = Scheduler()
    worker.start()
    for i in range(7):
        worker.append(f"sleep {i*2}")
    time.sleep(1)
    worker.stop()



import jammy
from jammy import io

parser = jammy.cli.JamArgumentParser()
parser.add_argument("filename", nargs="+")
args = parser.parse_args()


def simple():
    for i, filename in enumerate(args.filename):
        globals()[f"f{i + 1}"] = io.load(filename)

    from IPython import embed

    embed()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : backend.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import functools
import time

import numpy as np

from jammy.logging import get_logger

logger = get_logger()

try:
    import cv2
except ImportError:
    cv2 = None

try:
    from PIL import Image
except ImportError:
    Image = None

if cv2 is None:
    if Image is not None:
        logger.warning("Fail to import OpenCV; use PIL library.")
    else:
        logger.error(
            "Can not find either PIL OpenCV; you can not use most function in tartist.image."
        )


FORCE_PIL_BGR = False


def opencv_or_pil(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        if cv2 is None and Image is None:
            assert False, "Call {} without OpenCV or PIL.".format(func)
        return func(*args, **kwargs)

    return new_func


def opencv_only(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        if cv2 is None:
            assert False, "Call {} without OpenCV.".format(func)
        return func(*args, **kwargs)

    return new_func


def pil_only(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        if Image is None:
            assert False, "Call {} without PIL.".format(func)
        return func(*args, **kwargs)

    return new_func


def pil_img2nd(image, require_chl3=True):
    nd = np.array(image)
    if FORCE_PIL_BGR:
        nd = nd[:, :, ::-1]
    if require_chl3 and len(nd.shape) == 2:
        return nd[:, :, np.newaxis]
    return nd


def pil_nd2img(image):
    if FORCE_PIL_BGR:
        image = image[:, :, ::-1]
    if len(image.shape) == 3 and image.shape[2] == 1:
        image = image[:, :, 0]
    return Image.fromarray(image)


@opencv_or_pil
def imread(path):
    if cv2:
        # cv2 default BGR
        return cv2.imread(path)[..., ::-1]
    else:
        image = Image.open(path).convert("RGB")
        return pil_img2nd(image)


@opencv_or_pil
def imwrite(path, image):
    if cv2:
        return cv2.imwrite(path, image[..., ::-1])
    else:
        image = pil_nd2img(image)
        return image.save(path)


@opencv_or_pil
def imshow(title, image):
    if cv2:
        cv2.imshow(title, image)
        cv2.waitKey(0)
    else:
        image = pil_nd2img(image)
        image.show(title)
        time.sleep(0.5)


@opencv_or_pil
def resize(image, dsize, interpolation="LINEAR"):
    assert interpolation in ("NEAREST", "LINEAR", "CUBIC", "LANCZOS4")

    dsize = tuple(map(int, dsize))
    assert len(dsize) == 2
    if cv2:
        interpolation = getattr(cv2, "INTER_" + interpolation)
        return cv2.resize(image, dsize, interpolation=interpolation)
    else:
        if interpolation == "NEAREST":
            interpolation = Image.NEAREST
        elif interpolation == "LANCZOS4":
            interpolation = Image.LANCZOS
        else:
            interpolation = getattr(Image, "BI" + interpolation)

        image = pil_nd2img(image)
        image = image.resize(dsize, resample=interpolation)
        return pil_img2nd(image)



from jammy.cli.cmdline_viz import CmdLineViz

from . import progress_fn
from .genetic_trainer import GeneticTrainer
from .trainer_monitor import TrainerMonitor


class SimpleTrainer(GeneticTrainer):
    def __init__(self, cfg, loss_fn):
        super().__init__(cfg, loss_fn)
        self.cmdviz = CmdLineViz()
        self.trainer_monitor = None
        # cache state accross different methods
        self.cur_monitor = dict()
        self.cur_cmdviz = dict()
        self.set_up()

    def set_up(self):
        progress_fn.simple_train_bar(self)
        progress_fn.simple_val_bar(self)

    def monitor_update(self):
        if self.trainer_monitor:
            self.trainer_monitor.update(
                {
                    **self.cur_monitor,
                    "epoch": self.epoch_cnt,
                    "iter": self.iter_cnt,
                }
            )
        self.cur_monitor = dict()

    def set_monitor(self, is_wandb, tblogger=False):
        """
        docstring
        """
        self.trainer_monitor = TrainerMonitor(is_wandb, tblogger)



import inspect


def get_stack_frame(n=0, strict=True):
    """
    Gets the current stack frame or any of its ancestors dynamically
    Args:
        n (int): n=0 means the frame you called this function in.
                 n=1 is the parent frame.
        strict (bool): (default = True)
    Returns:
        frame: frame_cur
    Example:
        >>> frame_cur = get_stack_frame(n=0)
        >>> print('frame_cur = %r' % (frame_cur,))
        >>> assert frame_cur.f_globals['frame_cur'] is frame_cur
    """
    frame_cur = inspect.currentframe()
    # Use n+1 to always skip the frame of this function
    for idx in range(n + 1):
        frame_next = frame_cur.f_back
        if frame_next is None:  # nocover
            if strict:
                raise AssertionError("Frame level %r is root" % idx)
            break
        frame_cur = frame_next
    return frame_cur


def get_parent_frame(n=0):
    """
    Returns the frame of that called you.
    This is equivalent to `get_stack_frame(n=1)`
    Args:
        n (int): n=0 means the frame you called this function in.
                 n=1 is the parent frame.
    Returns:
        frame: parent_frame
    Example:
        >>> root0 = get_stack_frame(n=0)
        >>> def foo():
        >>>     child = get_stack_frame(n=0)
        >>>     root1 = get_parent_frame(n=0)
        >>>     root2 = get_stack_frame(n=1)
        >>>     return child, root1, root2
        >>> # Note this wont work in IPython because several
        >>> # frames will be inserted between here and foo
        >>> child, root1, root2 = foo()
        >>> print('root0 = %r' % (root0,))
        >>> print('root1 = %r' % (root1,))
        >>> print('root2 = %r' % (root2,))
        >>> print('child = %r' % (child,))
        >>> assert root0 == root1
        >>> assert root1 == root2
        >>> assert child != root1
    """
    parent_frame = get_stack_frame(n=n + 2)
    return parent_frame



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : lmdb.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.
import os
import pickle

import lmdb

from jammy.utils.cache import cached_property
from jammy.utils.container import OrderedSet

from .kv import KVStoreBase

__all__ = ["LMDBKVStore"]

_loads = pickle.loads
_dumps = pickle.dumps

# pylint: disable=attribute-defined-outside-init
class LMDBKVStore(KVStoreBase):
    _key_charset = "utf8"
    _magic_key = b"__keys__"

    def __init__(self, lmdb_path, readonly=True, keys=None):
        super().__init__(readonly=readonly)
        self._open(lmdb_path, readonly=readonly, keys=keys)

    def _open(self, lmdb_path, readonly, keys):
        self._lmdb = lmdb.open(
            lmdb_path,
            subdir=os.path.isdir(lmdb_path),
            readonly=readonly,
            lock=False,
            readahead=False,
            map_size=1099511627776 * 2,
            max_readers=100,
        )
        self._lmdb_keys = None
        if keys is not None:
            self._lmdb_keys = OrderedSet(keys)
        self._is_dirty = False

    @cached_property
    def txn(self):
        return self._lmdb.begin(write=not self.readonly)

    def _has(self, key):
        self._load_lmdb_keys()
        return key in self._lmdb_keys

    def _get(self, key, default):
        value = self.txn.get(key.encode(self._key_charset), default=default)
        value = _loads(value)
        return value

    def _put(self, key, value, replace=False):
        self._is_dirty = True
        self._load_lmdb_keys()
        self._lmdb_keys.append(key)
        return self.txn.put(
            key.encode(self._key_charset), _dumps(value), overwrite=replace
        )

    def _erase(self, key):
        self._is_dirty = True
        self._load_lmdb_keys()
        self._lmdb_keys.remove(key)
        return self.txn.pop(key.encode(self._key_charset))

    def _transaction(self, *args, **kwargs):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_trace):
        if exc_type:
            self.txn.abort()
        else:
            self.txn.put(self._magic_key, _dumps(self._lmdb_keys.as_list()))
            self.txn.commit()

    def _keys(self):
        self._load_lmdb_keys(True)
        return self._lmdb_keys

    def _load_lmdb_keys(self, assert_exist=False):
        if self._lmdb_keys is None:
            keys = self.txn.get(self._magic_key, None)
            if assert_exist:
                assert keys is not None, "LMDBKVStore does not support __keys__ access"
            if keys is None:
                self._lmdb_keys = OrderedSet()
            else:
                self._lmdb_keys = OrderedSet(_loads(keys))



class COLORS:
    """[This class is used to color the bash shell by using {} {} {} with 'COLORS.{}, text, COLORS.END_TOKEN']
    """

    TRAIN_COLOR = "\033[0;92m"
    VAL_COLOR = "\033[0;94m"
    TEST_COLOR = "\033[0;93m"
    BEST_COLOR = "\033[0;92m"

    END_TOKEN = "\033[0m)"
    END_NO_TOKEN = "\033[0m"

    Black = "\033[0;30m"  # Black
    Red = "\033[0;31m"  # Red
    Green = "\033[0;32m"  # Green
    Yellow = "\033[0;33m"  # Yellow
    Blue = "\033[0;34m"  # Blue
    Purple = "\033[0;35m"  # Purple
    Cyan = "\033[0;36m"  # Cyan
    White = "\033[0;37m"  # White

    # Bold
    BBlack = "\033[1;30m"  # Black
    BRed = "\033[1;31m"  # Red
    BGreen = "\033[1;32m"  # Green
    BYellow = "\033[1;33m"  # Yellow
    BBlue = "\033[1;34m"  # Blue
    BPurple = "\033[1;35m"  # Purple
    BCyan = "\033[1;36m"  # Cyan
    BWhite = "\033[1;37m"  # White

    # Underline
    UBlack = "\033[4;30m"  # Black
    URed = "\033[4;31m"  # Red
    UGreen = "\033[4;32m"  # Green
    UYellow = "\033[4;33m"  # Yellow
    UBlue = "\033[4;34m"  # Blue
    UPurple = "\033[4;35m"  # Purple
    UCyan = "\033[4;36m"  # Cyan
    UWhite = "\033[4;37m"  # White

    # Background
    On_Black = "\033[40m"  # Black
    On_Red = "\033[41m"  # Red
    On_Green = "\033[42m"  # Green
    On_Yellow = "\033[43m"  # Yellow
    On_Blue = "\033[44m"  # Blue
    On_Purple = "\033[45m"  # Purple
    On_Cyan = "\033[46m"  # Cyan
    On_White = "\033[47m"  # White

    # High Intensty
    IBlack = "\033[0;90m"  # Black
    IRed = "\033[0;91m"  # Red
    IGreen = "\033[0;92m"  # Green
    IYellow = "\033[0;93m"  # Yellow
    IBlue = "\033[0;94m"  # Blue
    IPurple = "\033[0;95m"  # Purple
    ICyan = "\033[0;96m"  # Cyan
    IWhite = "\033[0;97m"  # White

    # Bold High Intensty
    BIBlack = "\033[1;90m"  # Black
    BIRed = "\033[1;91m"  # Red
    BIGreen = "\033[1;92m"  # Green
    BIYellow = "\033[1;93m"  # Yellow
    BIBlue = "\033[1;94m"  # Blue
    BIPurple = "\033[1;95m"  # Purple
    BICyan = "\033[1;96m"  # Cyan
    BIWhite = "\033[1;97m"  # White

    # High Intensty backgrounds
    On_IBlack = "\033[0;100m"  # Black
    On_IRed = "\033[0;101m"  # Red
    On_IGreen = "\033[0;102m"  # Green
    On_IYellow = "\033[0;103m"  # Yellow
    On_IBlue = "\033[0;104m"  # Blue
    On_IPurple = "\033[10;95m"  # Purple
    On_ICyan = "\033[0;106m"  # Cyan
    On_IWhite = "\033[0;107m"  # White



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : session.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 10/23/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import hashlib
import hmac
import uuid

__all__ = ["Session", "SessionManagerBase"]


class Session(dict):
    def __init__(self, session_manager, request_handler):
        self._session_manager = session_manager
        self._request_handler = request_handler

        session_id, session_data = self._session_manager.get(self._request_handler)
        self._identifier = session_id
        super().__init__(session_data)

    @property
    def identifier(self):
        return self._identifier

    def save(self):
        self._session_manager.set(self._request_handler, self.identifier, dict(self))


class SessionManagerBase:
    def __init__(self, secret):
        self._secret = secret
        super().__init__()

    @property
    def secret(self):
        return self._secret

    def new(self, request_handler):
        return Session(self, request_handler)

    def get(self, request_handler):
        raise NotImplementedError()

    def set(self, request_handler, session_id, data):
        raise NotImplementedError()

    def _generate_id(self):
        new_id = hashlib.sha256((self.secret + str(uuid.uuid4())).encode("utf-8"))
        return new_id.hexdigest()

    def _generate_hmac(self, session_id):
        if not isinstance(session_id, bytes):
            session_id = session_id.encode("utf-8")
        return hmac.new(
            session_id, self.secret.encode("utf-8"), hashlib.sha256
        ).hexdigest()



import collections
import contextlib
import functools
import os
import sys
import threading
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import torch
from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger

import jammy.io as jio
from jammy.logging.logger import LOG_FORMAT, add_relative_path
from jammy.utils.env import jam_getenv
from jammy.utils.registry import CallbackRegistry

# skip precommit check for debug statement
if jam_getenv("pdb", "ipdb").lower() == "ipdb":
    import ipdb as pdb  # noqa
else:
    import pudb as pdb  # noqa # pylint: disable=import-error

__all__ = [
    "hook_exception_ipdb",
    "unhook_exception_ipdb",
    "exception_hook",
    "decorate_exception_hook",
    "replayer",
]

# pylint: disable=invalid-name, redefined-builtin


def _custom_exception_hook(type, value, tb):
    if hasattr(sys, "ps1") or not sys.stderr.isatty():
        # we are in interactive mode or we don't have a tty-like
        # device, so we call the default hook
        sys.__excepthook__(type, value, tb)
    else:
        import traceback

        # we are NOT in interactive mode, print the exception...
        traceback.print_exception(type, value, tb)
        # ...then start the debugger in post-mortem mode.
        pdb.post_mortem(tb)


def hook_exception_ipdb():
    if not hasattr(_custom_exception_hook, "origin_hook"):
        _custom_exception_hook.origin_hook = sys.excepthook
        sys.excepthook = _custom_exception_hook


def unhook_exception_ipdb():
    assert hasattr(_custom_exception_hook, "origin_hook")
    sys.excepthook = _custom_exception_hook.origin_hook


@contextlib.contextmanager
def exception_hook(enable=True):
    if enable:
        hook_exception_ipdb()
        yield
        unhook_exception_ipdb()
    else:
        yield


def decorate_exception_hook(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        with exception_hook():
            return func(*args, **kwargs)

    return wrapped


def _TimeoutEnterIpdbThread(locals_, cv, timeout):
    del locals_
    with cv:
        if not cv.wait(timeout):

            pdb.set_trace()


@contextlib.contextmanager
def timeout_ipdb(locals_, timeout=3):
    cv = threading.Condition()
    thread = threading.Thread(
        target=_TimeoutEnterIpdbThread, args=(locals_, cv, timeout)
    )
    thread.start()
    yield
    with cv:
        cv.notify_all()


def _default_replayer_fallback(obj, *args, **kwargs):
    del args, kwargs
    if len(obj) == 0:
        return obj
    item_type = type(obj)
    raise ValueError('Unknown itme type: "{}".'.format(item_type))


class _RePlayerCallbackRegistry(CallbackRegistry):
    def dispatch(self, name, *args, **kwargs):
        return super().dispatch_direct(name, *args, **kwargs)


def replay_proc_dict(obj: Dict[str, Any]):
    rtn_dict = {}
    for k, v in obj.items():
        rtn_dict[k] = replayer_processing(v)
    return rtn_dict


def replay_proc_module(obj: torch.nn.Module):
    grads_dict = {}
    for k, v in obj.named_parameters():
        # record weights value and grad if exists
        if v.grad is not None:
            grads_dict[k] = v.grad.clone().detach()
    return {
        "weights": obj.state_dict(),
        "grads": grads_dict,
    }


def replay_proc_basic(obj: List[Union[int, float, str, complex]]):
    assert type(obj) in [int, float, str, complex, np.ndarray, torch.Tensor]
    return obj


def replay_proc_iter(obj: Union[List, Tuple]):
    rtn = []
    for cur_obj in obj:
        rtn.append(replayer_processing(cur_obj))
    return rtn


_replayer_registry = _RePlayerCallbackRegistry()
_replayer_registry.set_fallback_callback(_default_replayer_fallback)
_replayer_registry.register(int, replay_proc_basic)
_replayer_registry.register(str, replay_proc_basic)
_replayer_registry.register(float, replay_proc_basic)
_replayer_registry.register(complex, replay_proc_basic)
_replayer_registry.register(np.ndarray, replay_proc_basic)
_replayer_registry.register(torch.Tensor, replay_proc_basic)
_replayer_registry.register(dict, replay_proc_dict)
_replayer_registry.register(collections.OrderedDict, replay_proc_dict)
_replayer_registry.register(collections.defaultdict, replay_proc_dict)
_replayer_registry.register(tuple, replay_proc_iter)
_replayer_registry.register(list, replay_proc_iter)


def replayer_processing(obj: Any):
    return _replayer_registry.dispatch(type(obj), obj)


class RePlayer:
    def __init__(self, savedir, level="INFO"):
        self.savedir = savedir
        self.counter = 0
        self.data = {}
        if not os.path.exists(savedir):
            os.makedirs(savedir)
        self.logger = _Logger(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=False,
            raw=False,
            capture=True,
            patcher=add_relative_path,
            extra={},
        )
        self.logger.add(
            os.path.join(savedir, "_jam_replayer.log"),
            level=level,
            mode="w",
            format=LOG_FORMAT,
        )
        self.logger.add(
            sys.stdout,
            level="INFO",
            format=LOG_FORMAT,
        )

    def info(self, *args, **kwargs):
        self.logger.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self.logger.debug(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self.logger.critical(*args, **kwargs)

    def log(self, **kwargs):
        self.data.update(replayer_processing(kwargs))

    def reset(self):
        self.data = {}
        self.counter = 0

    def dump(self):
        self.counter += 1
        jio.dump(
            f"{self.savedir}/jam_replay_{self.counter:04d}.pkl",
            self.data,
        )
        self.data = {}

    def load(self, idx, savedir=None):
        if savedir is None:
            savedir = self.savedir
        self.data = jio.load(
            f"{savedir}/jam_replay_{idx:04d}.pkl",
        )
        self.counter = idx


replayer = RePlayer("/tmp/jam_replayer")



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : container.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/18/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.

import copy
import collections

from .printing import kvformat, kvprint

__all__ = ["G", "g", "GView", "SlotAttrObject", "OrderedSet"]


class G(dict):
    def __getattr__(self, k):
        if k not in self:
            raise AttributeError
        return self[k]

    def __setattr__(self, k, v):
        self[k] = v

    def __delattr__(self, k):
        del self[k]

    def format(self, sep=": ", end="\n"):
        return kvformat(self, sep=sep, end=end)

    def print(self, sep=": ", end="\n", file=None):
        return kvprint(self, sep=sep, end=end, file=file)


g = G()


class GView(object):
    def __init__(self, dict_=None):
        if dict_ is None:
            dict_ = dict()
        object.__setattr__(self, "_dict", dict_)

    def __getattr__(self, k):
        if k not in self.raw():
            raise AttributeError
        return self.raw()[k]

    def __setattr__(self, k, v):
        self.raw()[k] = v

    def __delattr__(self, k):
        del self.raw()[k]

    def __getitem__(self, k):
        return self.raw()[k]

    def __setitem__(self, k, v):
        self.raw()[k] = v

    def __delitem__(self, k):
        del self.raw()[k]

    def __contains__(self, k):
        return k in self.raw()

    def __iter__(self):
        return iter(self.raw().items())

    def raw(self):
        return object.__getattribute__(self, "_dict")

    def update(self, other):
        self.raw().update(other)

    def copy(self):
        return GView(self.raw().copy())

    def format(self, sep=": ", end="\n"):
        return kvformat(self.raw(), sep=sep, end=end)

    def print(self, sep=": ", end="\n", file=None):
        return kvprint(self.raw(), sep=sep, end=end, file=file)


class SlotAttrObject(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __setattr__(self, k, v):
        assert not k.startswith("_")
        if k not in type(self).__dict__:
            # do not use hasattr; since it may result in infinite recursion
            raise AttributeError(
                "{}: could not set non-existing attribute {}".format(self, k)
            )
        cvt = getattr(type(self), "_convert_{}".format(k), None)
        if cvt is not None:
            v = cvt(v)
        super().__setattr__(k, v)

    def clone(self):
        return copy.deepcopy(self)


class OrderedSet(object):
    def __init__(self, initial_list=None):
        if initial_list is not None:
            self._dict = collections.OrderedDict([(v, True) for v in initial_list])
        else:
            self._dict = collections.OrderedDict()

    def append(self, value):
        self._dict[value] = True

    def remove(self, value):
        del self._dict[value]

    def __contains__(self, value):
        return value in self._dict

    def __iter__(self):
        return self._dict.keys()

    def as_list(self):
        return list(self._dict.keys())



from collections.abc import Iterable

import numpy as np
import torch
from einops import rearrange
from torch import nn

__all__ = ["FourierMLP"]


def check_shape(cur_shape):
    if isinstance(cur_shape, Iterable):
        return tuple(cur_shape)
    elif isinstance(cur_shape, int):
        return tuple(
            [
                cur_shape,
            ]
        )
    else:
        raise NotImplementedError(f"Type {type(cur_shape)} not support")


class FourierMLP(nn.Module):
    def __init__(self, in_shape, out_shape, num_layers=2, channels=128, zero_init=True):
        super().__init__()
        self.in_shape = check_shape(in_shape)
        self.out_shape = check_shape(out_shape)

        self.register_buffer(
            "timestep_coeff", torch.linspace(start=0.1, end=100, steps=channels)[None]
        )
        self.timestep_phase = nn.Parameter(torch.randn(channels)[None])
        self.input_embed = nn.Linear(int(np.prod(in_shape)), channels)
        self.timestep_embed = nn.Sequential(
            nn.Linear(2 * channels, channels),
            nn.GELU(),
            nn.Linear(channels, channels),
        )
        self.layers = nn.Sequential(
            nn.GELU(),
            *[
                nn.Sequential(nn.Linear(channels, channels), nn.GELU())
                for _ in range(num_layers)
            ],
            nn.Linear(channels, int(np.prod(self.out_shape))),
        )
        if zero_init:
            self.layers[-1].weight.data.fill_(0.0)
            self.layers[-1].bias.data.fill_(0.0)

    def forward(self, cond, inputs):
        cond = cond.view(-1, 1).expand((inputs.shape[0], 1))
        sin_embed_cond = torch.sin(
            (self.timestep_coeff * cond.float()) + self.timestep_phase
        )
        cos_embed_cond = torch.cos(
            (self.timestep_coeff * cond.float()) + self.timestep_phase
        )
        embed_cond = self.timestep_embed(
            rearrange([sin_embed_cond, cos_embed_cond], "d b w -> b (d w)")
        )
        embed_ins = self.input_embed(inputs.view(inputs.shape[0], -1))
        out = self.layers(embed_ins + embed_cond)
        return out.view(-1, *self.out_shape)


if __name__ == "__main__":
    from torchinfo import summary

    for t_in_shape, t_out_shape in [[3, 1], [3, 5], [(3,), 1], [(3, 5), (7, 3)]]:
        net = FourierMLP(t_in_shape, t_out_shape, 2)
        batch_size = 13
        design_in_shpae = check_shape(t_in_shape)
        summary(net, input_size=[(1,), (batch_size, *design_in_shpae)])
        summary(net, input_size=[(batch_size,), (batch_size, *design_in_shpae)])



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : packing.py
# Author : Jiayuan Mao, Qsh.Zh
# Email  : qsh.zh27@gmail.com
# Date   : 11/17/2021
#
# Qinsheng modifies based on Jacinle.
# Distributed under terms of the MIT license.

import functools
import os

from jammy.utils.enum import JamEnum
from jammy.utils.registry import CallbackRegistry, RegistryGroup

__all__ = [
    "check_pickle",
    "loadb_pickle",
    "dumpb_pickle",
    "check_msgpack",
    "loadb_msgpack",
    "dumpb_msgpack",
    "check_pyarrow",
    "loadb_pyarrow",
    "dumpb_pyarrow",
    "loadb",
    "dumpb",
    "get_available_backends",
    "get_default_backend",
    "set_default_backend",
]

import pickle

loadb_pickle = pickle.loads
dumpb_pickle = pickle.dumps

try:
    import msgpack
    import msgpack_numpy

    msgpack_numpy.patch()
    dumpb_msgpack = functools.partial(msgpack.dumps, use_bin_type=True)
    loadb_msgpack = msgpack.loads
except ImportError:
    dumpb_msgpack = loadb_msgpack = None

# pylint: disable=unnecessary-lambda, invalid-envvar-default

try:
    import pyarrow

    dumpb_pyarrow = lambda obj: pyarrow.serialize(obj).to_buffer()
    loadb_pyarrow = lambda buffer: pyarrow.deserialize(buffer)
except ImportError:
    dumpb_pyarrow = loadb_pyarrow = None


class _PackingFunctionRegistryGroup(RegistryGroup):
    __base_class__ = CallbackRegistry

    def dispatch(self, registry_name, entry, *args, **kwargs):
        return self[registry_name].dispatch(entry, *args, **kwargs)


_packing_function_registry = _PackingFunctionRegistryGroup()


def check_pickle():
    return True


def check_msgpack():
    return dumpb_msgpack is not None


def check_pyarrow():
    return dumpb_pyarrow is not None


class _PackingBackend(JamEnum):
    PICKLE = "pickle"
    MSGPACK = "msgpack"
    PYARROW = "pyarrow"


_packing_function_registry.register("check", _PackingBackend.PICKLE, lambda: True)
_packing_function_registry.register("check", _PackingBackend.MSGPACK, check_msgpack)
_packing_function_registry.register("check", _PackingBackend.PYARROW, check_pyarrow)


_packing_function_registry.register("loadb", _PackingBackend.PICKLE, loadb_pickle)
_packing_function_registry.register("dumpb", _PackingBackend.PICKLE, dumpb_pickle)

_packing_function_registry.register("loadb", _PackingBackend.MSGPACK, loadb_msgpack)
_packing_function_registry.register("dumpb", _PackingBackend.MSGPACK, dumpb_msgpack)

_packing_function_registry.register("loadb", _PackingBackend.PYARROW, loadb_pyarrow)
_packing_function_registry.register("dumpb", _PackingBackend.PYARROW, dumpb_pyarrow)

_default_packing_backend = _PackingBackend.PICKLE


def get_default_backend():
    return _default_packing_backend.name


def get_available_backends():
    return [
        obj.name
        for obj in _PackingBackend.choice_objs()
        if _packing_function_registry.dispatch("check", obj)
    ]


def set_default_backend(backend):
    global _default_packing_backend  # pylint: disable=global-statement
    _default_packing_backend = _PackingBackend.from_string(backend)
    assert (
        _default_packing_backend.name in get_available_backends()
    ), 'Unsupported backend on your machine: "{}".'.format(
        _default_packing_backend.name
    )


def loadb(bstr, *args, backend=None, **kwargs):
    backend = backend or _default_packing_backend
    return _packing_function_registry.dispatch("loadb", backend, bstr, *args, **kwargs)


def dumpb(obj, *args, backend=None, **kwargs):
    backend = backend or _default_packing_backend
    return _packing_function_registry.dispatch("dumpb", backend, obj, *args, **kwargs)


def _initialize_backend():
    set_default_backend(os.getenv("JAM_PACKING_BACKEND", _PackingBackend.PICKLE))


_initialize_backend()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : printing.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/18/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.

import collections
import contextlib
import io
import os
import sys
import threading

import numpy as np

from .registry import LockRegistry

__all__ = [
    "indent_text",
    "stprint",
    "stformat",
    "kvprint",
    "kvformat",
    "PrintToStringContext",
    "print_to_string",
    "print_to",
    "print2format",
    "bubbletext",
    "SupressPrintCTX",
]


_DEFAULT_FLOAT_FORMAT = "{:.6f}"


def indent_text(text, level=1, indent_format=None, tabsize=None):
    if indent_format is not None:
        assert tabsize is None, "Cannot provide both indent format and tabsize."
    if tabsize is not None:
        assert indent_format is None, "Cannot provide both indent format and tabsize."
        indent_format = " " * tabsize
    if indent_format is None and tabsize is None:
        indent_format = "  "
    indent_format = indent_format * level
    return indent_format + text.replace("\n", "\n" + indent_format)


def format_printable_data(data, float_format=_DEFAULT_FLOAT_FORMAT):
    t = type(data)
    if t is np.ndarray:
        return "ndarray{}, dtype={}".format(data.shape, data.dtype)
    # Handle torch.tensor
    if "Tensor" in str(t):
        return "tensor{}, dtype={}".format(tuple(data.shape), data.dtype)
    elif t is float:
        return float_format.format(data)
    else:
        return str(data)


def stprint(  # pylint: disable=too-many-arguments
    data,
    key=None,
    indent=0,
    file=None,
    indent_format="  ",
    end_format="\n",
    float_format=_DEFAULT_FLOAT_FORMAT,
    need_lock=True,
    max_depth=100,
):
    """
    Structure print.

    Example:

        >>> data = dict(a=np.zeros(shape=(10, 10)), b=3)
        >>> stprint(data)
        dict{
            a: ndarray(10, 10), dtype=float64
            b: 3
        }

    Args:
        data: data to be print. Currently support Sequnce, Mappings and primitive types.
        key: for recursion calls. Do not use it if you don't know how it works.
        indent: indent level.
    """

    if file is None:
        file = sys.stdout

    def _indent_print(msg, indent, prefix=None):
        print(indent_format * indent, end="", file=file)
        if prefix is not None:
            print(prefix, end="", file=file)
        print(msg, end=end_format, file=file)

    def _inner(data, indent, key, max_depth):
        t = type(data)
        if t is tuple:
            if max_depth == 0:
                _indent_print(
                    "(tuple of length {}) ...".format(len(data)), indent, prefix=key
                )
                return
            _indent_print("tuple[", indent, prefix=key)
            for v in data:
                _inner(v, indent=indent + 1, key=None, max_depth=max_depth - 1)
            _indent_print("]", indent)
        elif t is list:
            if max_depth == 0:
                _indent_print(
                    "(list of length {}) ...".format(len(data)), indent, prefix=key
                )
                return
            _indent_print("list[", indent, prefix=key)
            for v in data:
                _inner(v, indent=indent + 1, key=None, max_depth=max_depth - 1)
            _indent_print("]", indent)
        elif t in (dict, collections.OrderedDict, collections.defaultdict):
            if max_depth == 0:
                _indent_print(
                    "(dict of length {}) ...".format(len(data)), indent, prefix=key
                )
                return
            typename = {
                dict: "dict",
                collections.OrderedDict: "ordered_dict",
                collections.defaultdict: "default_dict",
            }[t]
            keys = sorted(data.keys())
            _indent_print(typename + "{", indent, prefix=key)
            for k in keys:
                v = data[k]
                _inner(
                    v, indent=indent + 1, key="{}: ".format(k), max_depth=max_depth - 1
                )
            _indent_print("}", indent)
        else:
            _indent_print(
                format_printable_data(data, float_format=float_format),
                indent,
                prefix=key,
            )

    with stprint.locks.synchronized(file, need_lock):
        _inner(data, indent=indent, key=key, max_depth=max_depth)

    del _inner


stprint.locks = LockRegistry()


def stformat(data, key=None, indent=0, max_depth=100, **kwargs):
    return print2format(stprint)(
        data, key=key, indent=indent, need_lock=False, max_depth=max_depth, **kwargs
    )


def kvprint(  # pylint: disable=too-many-arguments
    data,
    indent=0,
    sep=" : ",
    end="\n",
    max_key_len=None,
    file=None,
    float_format=_DEFAULT_FLOAT_FORMAT,
    need_lock=True,
):
    if len(data) == 0:
        return
    with kvprint.locks.synchronized(file, need_lock):
        keys = sorted(data.keys())
        lens = list(map(len, keys))
        if max_key_len is not None:
            max_len = max_key_len
        else:
            max_len = max(lens)
        for k in keys:
            print("  " * indent, end="")
            print(
                k + " " * max(max_len - len(k), 0),
                format_printable_data(data[k], float_format=float_format),
                sep=sep,
                end=end,
                file=file,
                flush=True,
            )


kvprint.locks = LockRegistry()


def kvformat(data, indent=0, sep=" : ", end="\n", max_key_len=None):
    return print2format(kvprint)(
        data, indent=indent, sep=sep, end=end, max_key_len=max_key_len, need_lock=False
    )


class PrintToStringContext:
    __global_locks = LockRegistry()

    def __init__(self, target="STDOUT", stream=None, need_lock=True):
        assert target in ("STDOUT", "STDERR")
        self._target = target
        self._need_lock = need_lock
        if stream is None:
            self._stream = io.StringIO()
        else:
            self._stream = stream
        self._stream_lock = threading.Lock()
        self._backup = None
        self._value = None

    def _swap(self, rhs):
        if self._target == "STDOUT":
            sys.stdout, rhs = rhs, sys.stdout
        else:
            sys.stderr, rhs = rhs, sys.stderr

        return rhs

    def __enter__(self):
        if self._need_lock:
            self.__global_locks[self._target].acquire()
        self._backup = self._swap(self._stream)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stream = self._swap(self._backup)
        if self._need_lock:
            self.__global_locks[self._target].release()

    def _ensure_value(self):
        if self._value is None:
            self._value = self._stream.getvalue()
            self._stream.close()

    def get(self):
        self._ensure_value()
        return self._value


def print_to_string(target="STDOUT"):
    return PrintToStringContext(target, need_lock=True)


@contextlib.contextmanager
def print_to(print_func, target="STDOUT", rstrip=True):
    with PrintToStringContext(target, need_lock=True) as ctx:
        yield
    out_str = ctx.get()
    if rstrip:
        out_str = out_str.rstrip()
    print_func(out_str)


def print2format(print_func):
    def format_func(*args, **kwargs):
        f = io.StringIO()
        print_func(*args, file=f, **kwargs)
        value = f.getvalue()
        f.close()
        return value

    return format_func


def bubbletext(text, font="cybermedium"):
    """
    Uses pyfiglet to create bubble text.
    Args:
        font (str): default=cybermedium, other fonts include: cybersmall and
            cyberlarge.
    References:
        http://www.figlet.org/
    Example:
        >>> bubble_text = bubbletext('TESTING BUBBLE TEXT', font='cybermedium')
        >>> print(bubble_text)
    """
    try:
        import pyfiglet
    except ImportError:
        from jammy.logging import get_logger

        logger = get_logger()
        logger.debug("Missing pyfiglet when use bubbletext")
        return text
    else:
        bubble_text = pyfiglet.figlet_format(text, font=font)
        return bubble_text


class SupressPrintCTX:
    def __init__(self, stdout=False, stderr=False):
        self._stdout = stdout
        self._stderr = stderr
        self._stdout_backup = None
        self._stderr_backup = None

    def __enter__(self):
        if self._stdout:
            self._stdout_backup = sys.stdout
            sys.stdout = open(os.devnull, "w")  # pylint: disable=unspecified-encoding
        if self._stderr:
            self._stderr_backup = sys.stderr
            sys.stderr = open(os.devnull, "w")  # pylint: disable=unspecified-encoding

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._stdout:
            sys.stdout.close()
            sys.stdout = self._stdout_backup
        if self._stderr:
            sys.stderr.close()
            sys.stderr = self._stderr_backup



import numpy as np
import torch as th
from torch.utils.data import Dataset

__all__ = ["TensorDataset"]


class TensorDataset(Dataset):
    def __init__(self, tensor):
        if isinstance(tensor, np.ndarray):
            data = th.from_numpy(tensor)
        elif isinstance(tensor, th.Tensor):
            data = tensor
        else:
            raise RuntimeError(f"{type(tensor)} not supported yet")

        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]



from typing import Tuple

import torch as th


class TorchRingBuffer:
    def __init__(self, item_shape: Tuple, max_size: int = 10000):
        self.max_size = max_size
        self._data = th.zeros((max_size, *item_shape))
        self.ptr_cnt = 0

    def add_item(self, item):
        self._data[self.ptr_cnt] = item
        self.ptr_cnt = (self.ptr_cnt + 1) % self.max_size

    def add_batch(self, batch_item):
        length = batch_item.shape[0]
        end_length = length + self.ptr_cnt
        if end_length > self.max_size:
            cut = end_length - self.max_size
            self._data[self.ptr_cnt :] = batch_item[:-cut]
            self._data[:cut] = batch_item[-cut:]
            self.ptr_cnt = cut
        else:
            self._data[self.ptr_cnt : end_length] = batch_item
            self.ptr_cnt = end_length

    def __getitem__(self, key):
        return self._data[key]

    @property
    def data(self):
        return self._data

    def to(self, device):  # pylint: disable=invalid-name
        self._data = self._data.to(device)


if __name__ == "__main__":
    buffer = TorchRingBuffer((3,), max_size=10)
    data = th.arange(15).view(-1, 1).expand((15, 3))
    buffer.add_batch(data)
    print(buffer.data)
    print(buffer[:])
    print(buffer[[1, 2, 3]])



def register_rng():
    from jammy.random.rng import global_rng_registry

    try:
        import torch
        # This will also automatically initialize cuda seeds.
        global_rng_registry.register('torch.manual_seed', lambda: torch.manual_seed)
        global_rng_registry.register('torch.cuda.manual_seed', lambda: torch.cuda.manual_seed)
        global_rng_registry.register('torch.cuda.manual_seed_all', lambda: torch.cuda.manual_seed_all)
    except ImportError:
        pass


def init_main():
    register_rng()


import matplotlib.pyplot as plt

__all__ = ["jcolors"]

# jcolors = ["blue", "orange", "green", "red", "purple", "cyan", "brown"]
# https://stackoverflow.com/questions/42086276/get-default-line-colour-cycle
jcolors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

plt_colors_names = [
    "aqua",
    "black",
    "blue",
    "blueviolet",
    "brown",
    "burlywood",
    "cadetblue",
    "coral",
    "cornflowerblue",
    "crimson",
    "darkblue",
    "darkgreen",
    "darkorange",
    "darkslateblue",
    "darkslategray",
    "darkviolet",
    "deeppink",
    "deepskyblue",
    "firebrick",
    "forestgreen",
    "fuchsia",
    "gray",
    "green",
    "hotpink",
    "lime",
    "maroon",
    "mediumorchid",
    "mediumpurple",
    "mediumseagreen",
    "mediumslateblue",
    "mediumspringgreen",
    "mediumturquoise",
    "mediumvioletred",
    "midnightblue",
    "mintcream",
    "mistyrose",
    "moccasin",
    "navajowhite",
    "navy",
    "oldlace",
    "olive",
    "olivedrab",
    "orangered",
    "orchid",
    "palegoldenrod",
    "palegreen",
    "paleturquoise",
    "palevioletred",
    "papayawhip",
    "peachpuff",
    "peru",
    "pink",
    "plum",
    "powderblue",
    "purple",
    "red",
    "rosybrown",
    "royalblue",
    "saddlebrown",
    "salmon",
    "sandybrown",
    "seagreen",
    "seashell",
    "sienna",
    "silver",
    "skyblue",
    "slateblue",
    "slategray",
    "snow",
    "springgreen",
    "steelblue",
    "tan",
    "teal",
    "thistle",
    "tomato",
    "turquoise",
    "violet",
    "wheat",
    "white",
    "whitesmoke",
    "yellow",
    "yellowgreen",
]


if __name__ == "__main__":
    import matplotlib.patches as mpatch

    num_color = 30
    fig, ax = plt.subplots(1, 1, figsize=(7 * 1, 17 * 1))

    n_rows = 1
    height = 1

    for j, color_name in enumerate(plt_colors_names[:num_color]):
        text_args = dict(fontsize=10, weight="bold")
        ax.add_patch(mpatch.Rectangle((0, j * height), 10, 1, color=color_name))
        ax.text(
            0.5 + 1,
            j * height + 0.7,
            color_name,
            color=color_name,
            ha="center",
            **text_args
        )

    # ax.set_xlim(0, 3 * n_groups)
    ax.set_ylim(num_color, -1)
    ax.axis("off")

    plt.show()



from tqdm.auto import tqdm
import collections
from tqdm import std
import numpy as np

from jammy.cli.colors import COLORS

from jammy.utils.meter import GroupMeters


class Coloredtqdm(tqdm):
    def __init__(self, *kargs, **kwargs):
        super(Coloredtqdm, self).__init__(*kargs, **kwargs)
        self._meter = GroupMeters()

    def set_postfix(
        self, ordered_dict=None, refresh=True, color=True, round=4, **kwargs
    ):
        postfix = std._OrderedDict([] if ordered_dict is None else ordered_dict)
        value_color = collections.defaultdict(lambda: COLORS.White)
        prev_read = self._meter.avg

        for key in sorted(kwargs.keys()):
            postfix[key] = kwargs[key]

        for key in postfix.keys():
            if key in prev_read:
                if prev_read[key] > postfix[key]:
                    value_color[key] = COLORS.Red
                else:
                    value_color[key] = COLORS.Green

        self._meter.update(postfix)

        for key in postfix.keys():
            if isinstance(postfix[key], std.Number):
                postfix[key] = self.format_num_to_k(
                    np.round(postfix[key], round), k=round + 1
                )
            if isinstance(postfix[key], std._basestring):
                postfix[key] = str(postfix[key])
            if len(postfix[key]) != round:
                postfix[key] += (round - len(postfix[key])) * " "

        if color:
            self.postfix = ", ".join(
                value_color[key] + key + "=" + postfix[key] + COLORS.END_NO_TOKEN
                for key in postfix.keys()
            )
        else:
            self.postfix = ", ".join(key + "=" + postfix[key] for key in postfix.keys())

        if refresh:
            self.refresh()

    def format_num_to_k(self, seq, k=4):
        seq = str(seq)
        length = len(seq)
        out = seq + " " * (k - length) if length < k else seq
        return out if length < k else seq[:k]



from .colors import COLORS
from tqdm.auto import tqdm

from collections import defaultdict

from jammy.utils.meter import GroupMeters


class CmdLineViz:
    def __init__(self):
        self.meter = defaultdict(GroupMeters)
        self.prev_mean = {}

    def update(self, mode, eval_dict):
        self.meter[mode].update(eval_dict)

    def flush(self):
        if len(self.meter) == 0:
            return

        _str = "\n" + "\t" * 1 + "=== Summary ===\n"
        keys = []
        color_write = {}
        for mode, meters in self.meter.items():
            avg = meters.avg
            color_write[mode] = defaultdict(lambda: " " * 10)
            keys.extend(list(avg.keys()))
            for key, value in avg.items():
                color = COLORS.White
                if mode + key in self.prev_mean:
                    color = (
                        COLORS.Green
                        if value > self.prev_mean[mode + key]
                        else COLORS.Red
                    )
                color_write[mode][key] = f"{color}{value:10.4f}{COLORS.END_NO_TOKEN}"
                self.prev_mean[mode + key] = value

        keys = set(keys)
        _str += f"\t{'mode':<10}"
        for cur_key in self.meter:
            _str += f" -- {cur_key:>10}"
        _str += "\n"
        for cur_key in keys:
            _str += f"\t{cur_key:10}"
            for _, writer in color_write.items():
                _str += f" -- {writer[cur_key]}"
            _str += f"\n"
        tqdm.write(_str + "\n")

        self.meter = defaultdict(GroupMeters)


if __name__ == "__main__":
    import numpy as np

    N = 10
    line = CmdLineViz()
    increase = np.arange(N) + np.random.randn(N)
    decrease = -np.arange(N) + np.random.randn(N)

    for cur_inc, cur_dec in zip(increase, decrease):
        line.update("test", {"increase": cur_inc, "decrease": cur_dec})

    line.flush()

    for cur_inc, cur_dec in zip(decrease, increase):
        line.update("test", {"increase": cur_inc, "decrease": cur_dec})

    line.flush()



from .meta import Singleton
from jammy.cli import timeout_input

__all__ = ["Gcfg", "pub_cfg", "set_pub_cfg", "get_pub_cfg", "omgcfg_create"]


class Gcfg(metaclass=Singleton):
    def __init__(self, cfg):
        assert "ready" not in cfg
        assert "cfg" not in cfg
        assert "reset" not in cfg
        assert "clear" not in cfg
        self.cfg = cfg
        for key, value in dict(cfg).items():
            self.__dict__[key] = value
            setattr(Gcfg, key, value)

    @classmethod
    def reset(cls, cfg=None):
        if cfg is None:
            cls._instances.clear()
        return Gcfg(cfg)


pub_cfg = None


def set_pub_cfg(cfg):
    global pub_cfg
    pub_cfg = cfg


def get_pub_cfg(cfg):
    return pub_cfg


def assert_name(cfg, default="Default", strict=True):
    if "name" not in cfg or cfg.name is None:
        cfg.name = timeout_input("cfg.name", 10, default, strict)
        return False
    return True


def omgcfg_create(obj):
    from omegaconf import OmegaConf

    return OmegaConf.create(obj)



import torch.nn as nn
from .base_modules import (
    BatchNorm1d,
    BatchNorm2d,
    BatchNorm3d,
    Conv1d,
    Conv2d,
    Conv3d,
    FC,
)

__all_ = ["Seq"]

class Seq(nn.Sequential):
    def __init__(self, input_channels):
        super(Seq, self).__init__()
        self.count = 0
        self.current_channels = input_channels

    def conv1d(
        self,
        out_size,
        kernel_size=1,
        stride=1,
        padding=0,
        dilation=1,
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=nn.init.kaiming_normal_,
        bias=True,
        preact=False,
        name="",
        norm_layer=BatchNorm1d,
    ):
        self.add_module(
            str(self.count),
            Conv1d(
                self.current_channels,
                out_size,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                activation=activation,
                bn=bn,
                init=init,
                bias=bias,
                preact=preact,
                name=name,
                norm_layer=norm_layer,
            ),
        )
        self.count += 1
        self.current_channels = out_size

        return self

    def conv2d(
        self,
        out_size,
        kernel_size=(1, 1),
        stride=(1, 1),
        padding=(0, 0),
        dilation=(1, 1),
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=nn.init.kaiming_normal_,
        bias=True,
        preact=False,
        name="",
        norm_layer=BatchNorm2d,
    ):

        self.add_module(
            str(self.count),
            Conv2d(
                self.current_channels,
                out_size,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                activation=activation,
                bn=bn,
                init=init,
                bias=bias,
                preact=preact,
                name=name,
                norm_layer=norm_layer,
            ),
        )
        self.count += 1
        self.current_channels = out_size

        return self

    def conv3d(
        self,
        out_size,
        kernel_size=(1, 1, 1),
        stride=(1, 1, 1),
        padding=(0, 0, 0),
        dilation=(1, 1, 1),
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=nn.init.kaiming_normal_,
        bias=True,
        preact=False,
        name="",
        norm_layer=BatchNorm3d,
    ):

        self.add_module(
            str(self.count),
            Conv3d(
                self.current_channels,
                out_size,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                activation=activation,
                bn=bn,
                init=init,
                bias=bias,
                preact=preact,
                name=name,
                norm_layer=norm_layer,
            ),
        )
        self.count += 1
        self.current_channels = out_size

        return self

    def fc(
        self,
        out_size,
        activation=nn.ReLU(inplace=True),
        bn=False,
        init=None,
        preact=False,
        name="",
    ):

        self.add_module(
            str(self.count),
            FC(
                self.current_channels,
                out_size,
                activation=activation,
                bn=bn,
                init=init,
                preact=preact,
                name=name,
            ),
        )
        self.count += 1
        self.current_channels = out_size

        return self

    def dropout(self, p=0.5):
        # type: (Seq, float) -> Seq

        self.add_module(str(self.count), nn.Dropout(p=0.5))
        self.count += 1

        return self

    def maxpool2d(
        self,
        kernel_size,
        stride=None,
        padding=0,
        dilation=1,
        return_indices=False,
        ceil_mode=False,
    ):
        self.add_module(
            str(self.count),
            nn.MaxPool2d(
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                return_indices=return_indices,
                ceil_mode=ceil_mode,
            ),
        )
        self.count += 1

        return self



# similar to https://github.com/invl/retry/blob/master/retry/api.py

import functools
import random
import time

from jammy.logging import get_logger

logging_logger = get_logger()

# pylint: disable=too-many-arguments, broad-except


def __retry_internal(
    f,
    exceptions=Exception,
    tries=-1,
    delay=0,
    max_delay=None,
    backoff=1,
    jitter=0,
    logger=logging_logger,
):
    """
    Executes a function and retries it if it failed.
    :param f: the function to execute.
    :param exceptions: an exception or a tuple of exceptions to catch. default: Exception.
    :param tries: the maximum number of attempts. default: -1 (infinite).
    :param delay: initial delay between attempts. default: 0.
    :param max_delay: the maximum value of delay. default: None (no limit).
    :param backoff: multiplier applied to delay between attempts. default: 1 (no backoff).
    :param jitter: extra seconds added to delay between attempts. default: 0.
                   fixed if a number, random if a range tuple (min, max)
    :param logger: logger.warning(fmt, error, delay) will be called on failed attempts.
                   default: retry.logging_logger. if None, logging is disabled.
    :returns: the result of the f function.
    """
    _tries, _delay = tries, delay
    while _tries:
        try:
            return f()
        except exceptions as e:
            _tries -= 1
            if not _tries:
                raise

            if logger is not None:
                logger.critical(f"{e}, retrying in {_delay} seconds...")

            time.sleep(_delay)
            _delay *= backoff

            if isinstance(jitter, tuple):
                _delay += random.uniform(*jitter)
            else:
                _delay += jitter

            if max_delay is not None:
                _delay = min(_delay, max_delay)


def retry_call(
    f,
    fargs=None,
    fkwargs=None,
    exceptions=Exception,
    tries=-1,
    delay=0,
    max_delay=None,
    backoff=1,
    jitter=0,
    logger=logging_logger,
):
    """
    Calls a function and re-executes it if it failed.
    :param f: the function to execute.
    :param fargs: the positional arguments of the function to execute.
    :param fkwargs: the named arguments of the function to execute.
    :param exceptions: an exception or a tuple of exceptions to catch. default: Exception.
    :param tries: the maximum number of attempts. default: -1 (infinite).
    :param delay: initial delay between attempts. default: 0.
    :param max_delay: the maximum value of delay. default: None (no limit).
    :param backoff: multiplier applied to delay between attempts. default: 1 (no backoff).
    :param jitter: extra seconds added to delay between attempts. default: 0.
                   fixed if a number, random if a range tuple (min, max)
    :param logger: logger.warning(fmt, error, delay) will be called on failed attempts.
                   default: retry.logging_logger. if None, logging is disabled.
    :returns: the result of the f function.
    """
    args = fargs if fargs else list()
    kwargs = fkwargs if fkwargs else dict()
    return __retry_internal(
        functools.partial(f, *args, **kwargs),
        exceptions,
        tries,
        delay,
        max_delay,
        backoff,
        jitter,
        logger,
    )


def retry(
    exceptions=Exception,
    tries=-1,
    delay=0,
    max_delay=None,
    backoff=1,
    jitter=0,
    logger=logging_logger,
):
    """Returns a retry decorator.
    :param exceptions: an exception or a tuple of exceptions to catch. default: Exception.
    :param tries: the maximum number of attempts. default: -1 (infinite).
    :param delay: initial delay between attempts. default: 0.
    :param max_delay: the maximum value of delay. default: None (no limit).
    :param backoff: multiplier applied to delay between attempts. default: 1 (no backoff).
    :param jitter: extra seconds added to delay between attempts. default: 0.
                   fixed if a number, random if a range tuple (min, max)
    :param logger: logger.warning(fmt, error, delay) will be called on failed attempts.
                   default: retry.logging_logger. if None, logging is disabled.
    :returns: a retry decorator.
    """

    @functools.wraps
    def retry_decorator(f, *fargs, **fkwargs):
        args = fargs if fargs else list()
        kwargs = fkwargs if fkwargs else dict()
        return __retry_internal(
            functools.partial(f, *args, **kwargs),
            exceptions,
            tries,
            delay,
            max_delay,
            backoff,
            jitter,
            logger,
        )

    return retry_decorator



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : application.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 10/23/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.
# pylint: skip-file
import time
from copy import deepcopy

from tornado.web import Application, RequestHandler

from jammy.logging import get_logger
from jammy.utils.imp import load_module

logger = get_logger()

__all__ = ["route", "make_app", "get_app", "JamApplication", "JamRequestHandler"]

__handlers__ = []
__application__ = None
__default_frontend_salt__ = (
    "|v<zj!#PGnxgc])Sd_=i|8,?,7><s5+Ck_IWcbC D:@4a~(1bEA=M+MacyI2udaz"
)

# pylint: disable=global-statement
def route(regex):
    global __handlers__

    def wrap_class(cls):
        __handlers__.append((regex, cls))
        return cls

    return wrap_class


def make_app(modules, settings):
    global __application__, __handlers__
    assert __application__ is None

    for m in modules:
        load_module(m)
    __application__ = JamApplication(__handlers__, **settings)
    return __application__


def get_app():
    global __application__
    return __application__


try:
    _ARG_DEFAULT = RequestHandler._ARG_DEFAULT
except:  # pylint: disable=bare-except
    from tornado.web import _ARG_DEFAULT


class JamRequestHandler(RequestHandler):
    def __init__(self, *args, **kwargs):
        self.session = None
        self.__start_time = None
        self.__end_time = None

        super().__init__(*args, **kwargs)

    def initialize(self):
        if self.application.session_enabled:
            self.session = self.application.get_session(self)
        self.__start_time = time.time()

    def save_session(self):
        if self.application.session_enabled:
            self.session.save()

    def get_template_namespace(self):
        dct = super().get_template_namespace()
        dct["url"] = self.request.uri
        dct["url_root"] = self.application.settings.get("url_root", "/")
        dct["is_localhost"] = self.application.settings.get("is_localhost", False)
        dct["frontend_secret"] = self.application.settings.get(
            "frontend_secret", __default_frontend_salt__
        )
        return dct

    def finish(self, *args, **kwargs):
        self.on_body_finish()
        super().finish(*args, **kwargs)

    def on_body_finish(self):
        self.save_session()

    def on_finish(self):
        self.__end_time = time.time()
        logger.info(
            "{} query {} executed for {:.3f} ms".format(
                self.request.remote_ip,
                self.request.uri,
                (self.__end_time - self.__start_time) * 1000,
            )
        )
        super().on_finish()

    def get_query_argument(
        self, name, default=_ARG_DEFAULT, strip=True, type=None, danger_set=None
    ):
        return self._get_argument(
            name, default, self.request.arguments, strip, type, danger_set
        )

    def get_body_argument(
        self, name, default=_ARG_DEFAULT, strip=True, type=None, danger_set=None
    ):
        return self._get_argument(
            name, default, self.request.body_arguments, strip, type, danger_set
        )

    def get_argument(
        self, name, default=_ARG_DEFAULT, strip=True, type=None, danger_set=None
    ):
        return self._get_argument(
            name, default, self.request.arguments, strip, type, danger_set
        )

    def _get_argument(
        self, name, default, source, strip=True, type=None, danger_set=None
    ):
        raw = super()._get_argument(name, default, source, strip)
        if type is int:
            return self._assert_int(raw)
        elif type is str:
            return self._assert_safe_string(raw, danger_set)
        elif type is None:
            return raw
        else:
            raise NotImplementedError("Unknown argument type: {}.".format(type))

    def _assert_int(self, v):
        try:
            v = int(v)
        except ValueError:
            raise JamRequestException("Invalid data submitted.")
        return v

    def _assert_safe_string(self, s, danger_set):
        if danger_set is None:
            danger_set = "`#'\"<>"
        try:
            s = str(s)
            for i in s:
                if i in danger_set:
                    raise ValueError
        except ValueError:
            raise JamRequestException("Invalid data submitted.")
        return s


class JamApplication(Application):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.settings["session_engine"] == "off":
            self.session_enabled = False
            self.session_manager = None
        elif self.settings["session_engine"] == "memcached":
            from jamweb.session.memcached import MemcachedSessionManager

            self.session_enabled = True
            self.session_manager = MemcachedSessionManager(
                secret=self.settings["session_secret"],
                memcache_host=self.settings["memcached_host"],
                memcache_port=self.settings["memcached_port"],
                timeout=self.settings["session_timeout"],
                cookie_prefix=self.settings["session_cookie_prefix"],
                memcached_prefix=self.settings["session_memcached_prefix"],
            )
            logger.critical(
                "Initializing the session manager using memcached: {}.".format(
                    self.session_manager.memcache.full_addr
                )
            )
        else:
            raise ValueError(
                "Unknown session engine: {}.".format(self.settings["session_engine"])
            )

    def get_session(self, request_handler):
        assert self.session_manager is not None, "Session manager not initialized."
        return self.session_manager.new(request_handler)


class JamRequestException(Exception):
    pass



from .pytorch_util import *
from .pytorch_util import set_gpu_mode as _set_gpu_mode
from .optim import *


def set_gpu_mode(is_gpu=True, gpu_id=0):
    """ this is to fix the buf import variable from file.
    set jam pytorch utils device

    :param is_gpu: [description], defaults to True
    :type is_gpu: bool, optional
    :param gpu_id: [description], defaults to 0
    :type gpu_id: int, optional
    """
    import torch

    global device
    device = torch.device("cuda:" + str(gpu_id) if is_gpu else "cpu")
    _set_gpu_mode(is_gpu, gpu_id)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : context.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 12/31/16
#
# Distributed under terms of the MIT license.

__all__ = ["EmptyContext"]


class EmptyContext(object):
    def __enter__(self):
        return

    def __exit__(self, exc_type, exc_val, exc_tb):
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : meter.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import six
import itertools
import collections

import jammy.io as io

from .meta import map_exec


class AverageMeter(object):
    """Computes and stores the average and current value"""

    val = 0
    avg = 0
    sum = 0
    count = 0
    tot_count = 0

    def __init__(self):
        self.reset()
        self.tot_count = 0

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.tot_count += n
        self.avg = self.sum / self.count


class GroupMeters(object):
    def __init__(self):
        self._meters = collections.defaultdict(AverageMeter)

    def reset(self):
        map_exec(AverageMeter.reset, self._meters.values())

    def update(self, updates=None, value=None, n=1, **kwargs):
        """
        Example:
            >>> meters.update(key, value)
            >>> meters.update({key1: value1, key2: value2})
            >>> meters.update(key1=value1, key2=value2)
        """
        if updates is None:
            updates = {}
        if updates is not None and value is not None:
            updates = {updates: value}
        updates.update(kwargs)
        for k, v in updates.items():
            self._meters[k].update(v, n=n)

    def __getitem__(self, name):
        return self._meters[name]

    def items(self):
        return self._meters.items()

    @property
    def sum(self):
        return {k: m.sum for k, m in self._meters.items() if m.count > 0}

    @property
    def avg(self):
        return {k: m.avg for k, m in self._meters.items() if m.count > 0}

    @property
    def val(self):
        return {k: m.val for k, m in self._meters.items() if m.count > 0}

    def format(self, caption, values, kv_format, glue):
        meters_kv = self._canonize_values(values)
        log_str = [caption]
        log_str.extend(itertools.starmap(kv_format.format, sorted(meters_kv.items())))
        return glue.join(log_str)

    def format_simple(self, caption, values="avg", compressed=True):
        if compressed:
            return self.format(caption, values, "{}={:4f}", " ")
        else:
            return self.format(caption, values, "\t{} = {:4f}", "\n")

    def dump(self, filename, values="avg"):
        meters_kv = self._canonize_values(values)
        with open(filename, "a") as f:
            f.write(io.dumps_json(meters_kv, compressed=False))
            f.write("\n")

    def _canonize_values(self, values):
        if isinstance(values, six.string_types):
            assert values in ("avg", "val", "sum")
            meters_kv = getattr(self, values)
        else:
            meters_kv = values
        return meters_kv



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : device.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/18/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import os
import collections

from jammy.utils.enum import JamEnum

__all__ = [
    "DeviceNameFormat",
    "canonlize_device_name",
    "parse_devices",
    "set_cuda_visible_devices",
    "parse_and_set_devices",
]


class DeviceNameFormat(JamEnum):
    TENSORFLOW = "tensorflow"
    INT = "int"


def canonlize_device_name(d, format=DeviceNameFormat.INT):
    format = DeviceNameFormat.from_string(format)

    d = d.strip().lower()
    if d == "cpu":
        if format is DeviceNameFormat.TENSORFLOW:
            return "/cpu:0"
        elif format is DeviceNameFormat.INT:
            return -1

    if d.startswith("gpu"):
        d = d[3:]
    d = int(d)

    if format is DeviceNameFormat.TENSORFLOW:
        return "/gpu:" + str(d)
    elif format is DeviceNameFormat.INT:
        return d


def parse_devices(devs, format=DeviceNameFormat.INT):
    if type(devs) is str:
        devs = devs.split(",")
    else:
        assert isinstance(devs, collections.Sequence)
        if len(devs) == 0:
            devs = devs[0].split(",")

    devs = [canonlize_device_name(d, format) for d in devs]
    return devs


def set_cuda_visible_devices(devs):
    devs = parse_devices(devs, DeviceNameFormat.INT)
    all_gpus = [str(d) for d in devs if d > -1]  # select only GPUs.
    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(all_gpus)


def parse_and_set_devices(devs, format=DeviceNameFormat.INT, set_device=True):
    if set_device:
        set_cuda_visible_devices(devs)
    return parse_devices(devs, format)



import inspect
import os
import os.path as osp
from collections.abc import Mapping
from functools import partial, update_wrapper

import hydra
from omegaconf import DictConfig, OmegaConf

from jammy import io
from jammy.logging import get_logger
from jammy.utils import imp

logger = get_logger()

__all__ = ["hydpath", "instantiate", "hyd_instantiate", "link_hyd_run", "update_cfg"]


def path(input_path=None):
    if input_path is None or len(input_path) == 0:
        return hydra.utils.to_absolute_path(".")
    if input_path[0] == "/":
        return input_path
    if input_path[0] == "~":
        return osp.expanduser(input_path)
    return hydra.utils.to_absolute_path(input_path)


def hydpath(input_path=None):
    return path(input_path)


def hyd_instantiate(_cfg, *args, **kwargs):
    """
    A helper func helps initialize from omegaconf(hydra) _cfg

    Args:
        _cfg: omegaconf.DictConfig, must contain `_target_` key word
             if `param` or `params` present in the _cfg, they are
             high priority to initialize instance
             otherwise, use _cfg

    Note:
        when use the method pay attention to the parameters order

    args and kwargs: other parameters needed to initialize network
    """
    if _cfg is None:
        stack = inspect.stack()[1]
        logger.info(
            f"File {stack.filename} {stack.lineno}: {stack.function} creating None object"
        )
        return None
    module = imp.load_class(_cfg["_target_"])
    if "params" in _cfg:
        _params = OmegaConf.to_container(_cfg.params, resolve=True)
    elif "param" in _cfg:
        _params = OmegaConf.to_container(_cfg.param, resolve=True)
    else:
        _params = OmegaConf.to_container(_cfg, resolve=True)
        del _params["_target_"]
    if inspect.isclass(module):
        return module(*args, **kwargs, **_params)
    if inspect.isfunction(module):
        partial_fn = partial(module, *args, **kwargs, **_params)
        update_wrapper(partial_fn, module)
        return partial_fn

    raise RuntimeError(" only support function and class")


def instantiate(_cfg, *args, **kwargs):
    """
    A helper func helps instantiate from omegaconf(hydra) _cfg

    Args:
        _cfg: omegaconf.DictConfig, must contain `_target_` key word
             if `param` or `params` present in the _cfg, they are
             high priority to initialize instance
             otherwise, use _cfg

    Note:
        when use the method pay attention to the parameters order

    args and kwargs: other parameters needed to initialize network
    """
    instance_ = hyd_instantiate(_cfg, *args, **kwargs)
    if inspect.isfunction(instance_):
        return instance_()
    if isinstance(instance_, partial):
        return instance_()
    return instance_


def flatten_dict(cfg):
    rtn = {}
    for k, v in cfg.items():
        if isinstance(v, Mapping):
            sub = {f"{k}.{sub_k}": sub_v for sub_k, sub_v in flatten_dict(v).items()}
            rtn.update(sub)
        else:
            rtn[k] = v
    return rtn


def link_hyd_run(dst_fname=".latest_exp", proj_path=None):
    exp_folder = os.getcwd()
    if proj_path is None:
        proj_path = hydra.utils.get_original_cwd()
    io.link(exp_folder, osp.join(proj_path, dst_fname), use_relative_path=False)
    logger.info(f"{exp_folder} ==>> {osp.join(proj_path, dst_fname)}")


def update_cfg(cfg: DictConfig, dotlist):
    new_conf = OmegaConf.from_dotlist(dotlist)
    return OmegaConf.merge(cfg, new_conf)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : registry.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import collections
import uuid

from jammy.utils.registry import Registry, RegistryGroup

__all__ = [
    "SimpleEventRegistry",
    "EventRegistry",
    "EventRegistryGroup",
    "register_event",
    "trigger_event",
]


class SimpleEventRegistry(object):
    def __init__(self, allowed_events=None):
        self._allowed_events = allowed_events
        self._events = collections.defaultdict(list)

    def register(self, event, callback):
        if self._allowed_events is not None:
            assert event in self._allowed_events
        self._events[event].append(callback)

    def trigger(self, event, *args, **kwargs):
        if self._allowed_events is not None:
            assert event in self._allowed_events
        for f in self._events[event]:
            f(*args, **kwargs)


class EventRegistry(Registry):
    DEF_PRIORITY = 10

    def _init_registry(self):
        self._registry = collections.defaultdict(  # entry
            lambda: collections.defaultdict(  # priority
                collections.OrderedDict  # sub-key
            )
        )

    def register(self, entry, callback, priority=DEF_PRIORITY, subkey=None):
        if subkey is None:
            subkey = uuid.uuid4()
        self._registry[entry][priority][subkey] = callback
        return subkey

    def unregister(self, entry, key=None, priority=DEF_PRIORITY):
        if key is None:
            return self._registry.pop(entry, None)

        entries = self._registry[entry][priority]
        if callable(entry):
            for k, v in list(entries.items()):
                return entries.pop(k)
            return None
        return entries.pop(key, None)

    def lookup(self, entry, key=None, priority=DEF_PRIORITY, default=None):
        if key is None:
            return self._registry.get(key, default)

        return self._registry[entry][priority].get(key, default)

    def trigger(self, entry, *args, **kwargs):
        self.trigger_args(entry, args, kwargs)

    def trigger_args(self, entry, args, kwargs):
        group = self._registry[entry]
        entry_lists = [group[k] for k in sorted(group.keys())]

        for entries in entry_lists:
            for entry in entries:
                entry(*args, **kwargs)


class EventRegistryGroup(RegistryGroup):
    __base_class__ = EventRegistry

    def lookup(self, registry_name, entry, key=None, **kwargs):
        return self._registries[registry_name].lookup(entry, key=key, **kwargs)

    def trigger(self, registry_name, entry, *args, **kwargs):
        return self._registries[registry_name].trigger(entry, *args, **kwargs)


default_event_registry_group = EventRegistryGroup()
register_event = default_event_registry_group.register
trigger_event = default_event_registry_group.trigger



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : zmq_utils.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/22/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import json
import socket
import uuid

import zmq

from jammy.concurrency.packing import dumpb, loadb

json_dumpb = lambda x: json.dumps(x).encode("utf-8")
json_loadb = lambda x: json.loads(x.decode("utf-8"))


def router_recv_json(sock, flag=zmq.NOBLOCK, loader=json_loadb):
    try:
        identifier, _, *payload = sock.recv_multipart(flag)
        return [identifier] + list(map(loader, payload))
    except zmq.error.ZMQError:
        return None, None


def router_send_json(sock, identifier, *payloads, flag=0, dumper=json_dumpb):
    try:
        buf = [identifier, b""]
        buf.extend(map(dumper, payloads))
        sock.send_multipart(buf, flags=flag)
    except zmq.error.ZMQError:
        return False
    return True


def req_recv_json(sock, flag=0, loader=json_loadb):
    try:
        response = sock.recv_multipart(flag)
        response = list(map(loader, response))
        return response[0] if len(response) == 1 else response
    except zmq.error.ZMQError:
        return None


def req_send_json(sock, *payloads, flag=0, dumper=json_dumpb):
    buf = []
    buf.extend(map(dumper, payloads))
    try:
        sock.send_multipart(buf, flag)
    except zmq.error.ZMQError:
        return False
    return True


def iter_recv(meth, sock):
    while True:
        res = meth(sock, flag=zmq.NOBLOCK)
        succ = res[0] is not None if isinstance(res, (tuple, list)) else res is not None
        if succ:
            yield res
        else:
            break


def req_send_and_recv(sock, *payloads):
    req_send_json(sock, *payloads)
    return req_recv_json(sock)


def push_pyobj(sock, data, flag=zmq.NOBLOCK):
    try:
        sock.send(dumpb(data), flag, copy=False)
    except zmq.error.ZMQError:
        return False
    return True


def pull_pyobj(sock, flag=zmq.NOBLOCK):
    try:
        response = loadb(sock.recv(flag, copy=False).bytes)
        return response
    except zmq.error.ZMQError:
        return None


def bind_to_random_ipc(sock, name):
    name = name + uuid.uuid4().hex[:8]
    conn = "ipc:///tmp/{}".format(name)
    sock.bind(conn)
    return conn


def uid():
    return socket.gethostname() + "/" + uuid.uuid4().hex


def graceful_close(sock):
    if sock is None:
        return
    sock.setsockopt(zmq.LINGER, 0)
    sock.close()



import jammy.logging as logging
from jamtorch.utils.meta import is_master

try:
    import torch_xla.core.xla_model as xm

    IS_TPU = True
except Exception as e:
    IS_TPU = False

__all__ = ["get_logger"]


def get_logger(*args, **kwargs):
    if IS_TPU:
        if xm.get_ordinal == 0:
            return logging.get_logger(*args, **kwargs)
    elif is_master():
        return logging.get_logger(*args, **kwargs)

    return logging.fake_logger



import sys

from jammy.utils.meter import AverageMeter
from jamtorch.utils.meta import as_float
from tqdm.auto import tqdm
from jammy.cli.cmdline_viz import CmdLineViz
from .monitor import group_prefix

__all__ = [
    "set_monitor_viz",
    "train_bar_init",
    "batch_bar_update",
    "epoch_bar_update",
    "batch_bar_new",
    "bar_close",
    "simple_step_summary",
    "val_start",
    "val_step",
    "val_end",
    "simple_train_bar",
    "simple_val_bar",
]


def set_monitor_viz(trainer):
    trainer.cur_monitor = dict()
    trainer.cmdviz = CmdLineViz()


def train_bar_init(trainer):
    trainer.num_batch = len(trainer.train_loader)
    trainer.epoch_bar = tqdm(
        total=trainer._cfg.epochs,
        desc="epochs",
        dynamic_ncols=True,
        initial=trainer.epoch_cnt,
        file=sys.stdout,
    )
    trainer.batch_bar = tqdm(
        total=trainer.num_batch,
        leave=False,
        desc="train",
        dynamic_ncols=True,
        file=sys.stdout,
    )


def batch_bar_update(trainer):
    trainer.batch_bar.update()
    trainer.batch_bar.set_postfix(
        dict(total_it=trainer.iter_cnt, loss=trainer.batch_loss)
    )


def epoch_bar_update(trainer):
    trainer.epoch_bar.update()
    trainer.epoch_bar.set_postfix(
        dict(total_it=trainer.iter_cnt, loss=trainer.batch_loss)
    )


def batch_bar_new(trainer):
    trainer.batch_bar.close()
    trainer.batch_bar = tqdm(
        total=trainer.num_batch,
        leave=False,
        desc="train",
        dynamic_ncols=True,
        file=sys.stdout,
    )


def bar_close(trainer):
    trainer.epoch_bar.close()
    trainer.batch_bar.close()
    sys.stdout.flush()


def val_start(trainer):
    trainer.val_loss_meter = AverageMeter()
    trainer.val_bar = tqdm(
        total=len(trainer.val_loader),
        leave=False,
        desc="val",
        dynamic_ncols=True,
        file=sys.stdout,
    )


def val_step(trainer, batch, loss, monitor, cmdviz_dict):
    trainer.val_loss_meter.update(loss)
    cmdviz_dict.update({"loss": trainer.val_loss_meter.val})
    f_cmdviz = as_float(cmdviz_dict)
    trainer.cmdviz.update("eval", f_cmdviz)

    trainer.val_bar.update()
    trainer.val_bar.set_postfix({"loss": trainer.val_loss_meter.avg.item()})


def val_end(trainer):
    val_loss = trainer.val_loss_meter.avg.item()
    trainer.cur_monitor.update(group_prefix("eval", trainer.cmdviz.meter["eval"].avg))
    trainer.val_bar.close()
    trainer.save_ckpt(val_loss)
    trainer.cmdviz.flush()


def simple_step_summary(trainer, loss, monitors, cmdviz_dict):
    loss_f = as_float(loss)
    cmdviz_f = as_float(cmdviz_dict)
    monitors = as_float(monitors)

    cmdviz_f.update({"loss": loss_f})
    monitors.update(cmdviz_f)

    trainer.batch_loss = loss_f
    trainer.cur_monitor.update(group_prefix("train", monitors))
    trainer.cmdviz.update("train", cmdviz_f)


def simple_train_bar(trainer):
    trainer.register_event("epoch:start", set_monitor_viz, False)
    trainer.register_event("epoch:start", train_bar_init, False)
    trainer.register_event("step:end", batch_bar_update, False)
    trainer.register_event("epoch:after", batch_bar_new, False)
    trainer.register_event("epoch:after", epoch_bar_update, False)
    trainer.register_event("epoch:end", bar_close, False)

    trainer.register_event("step:summary", simple_step_summary, False)


def simple_val_bar(trainer):
    trainer.register_event("epoch:start", set_monitor_viz, False)
    trainer.register_event("val:start", val_start, False)
    trainer.register_event("val:step", val_step, False)
    trainer.register_event("val:end", val_end, False)
